from elke27_lib.dispatcher import PagedBlock
from elke27_lib.handlers.zone import make_zone_configured_merge
from elke27_lib.states import PanelState


def test_zone_get_configured_paging_merge_accumulates_blocks():
    state = PanelState()
    merge = make_zone_configured_merge(state)

    blocks = [
        PagedBlock(
            block_id=1, payload={"block_id": 1, "block_count": 4, "zones": list(range(1, 17))}
        ),
        PagedBlock(
            block_id=2, payload={"block_id": 2, "block_count": 4, "zones": list(range(17, 33))}
        ),
        PagedBlock(
            block_id=3, payload={"block_id": 3, "block_count": 4, "zones": list(range(33, 49))}
        ),
        PagedBlock(block_id=4, payload={"block_id": 4, "block_count": 4, "zones": [49, 50]}),
    ]

    merged = merge(blocks, 4)

    assert merged["zones"] == list(range(1, 51))
    assert all(zone_id not in merged["zones"] for zone_id in range(51, 65))


def test_zone_get_configured_paging_merge_out_of_order_and_duplicates():
    state = PanelState()
    merge = make_zone_configured_merge(state)

    blocks = [
        PagedBlock(
            block_id=3, payload={"block_id": 3, "block_count": 4, "zones": list(range(33, 49))}
        ),
        PagedBlock(
            block_id=1, payload={"block_id": 1, "block_count": 4, "zones": list(range(1, 17))}
        ),
        PagedBlock(
            block_id=2, payload={"block_id": 2, "block_count": 4, "zones": list(range(17, 33))}
        ),
        PagedBlock(
            block_id=2, payload={"block_id": 2, "block_count": 4, "zones": list(range(17, 33))}
        ),
        PagedBlock(block_id=4, payload={"block_id": 4, "block_count": 4, "zones": [49, 50]}),
    ]

    merged = merge(blocks, 4)

    assert merged["zones"] == list(range(1, 51))



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import cast

_REQUIRED_KEYS: dict[str, tuple[str, ...]] = {
    "control_authenticate": (
        "seq",
        "session_id",
        "installer",
        "user_id",
        "group_id",
        "network_csm",
        "area_csm",
        "user_group_csm",
        "user_csm",
        "zone_csm",
        "keyfob_csm",
        "system_csm",
        "wltx_csm",
        "output_csm",
        "cs_param_csm",
        "flag_csm",
        "pir_flag_csm",
        "user_value_csm",
        "register_csm",
        "rule_csm",
        "task_csm",
        "light_csm",
        "lock_csm",
        "tstat_csm",
        "barrier_csm",
        "win_cover_csm",
        "timer_csm",
        "zwave_csm",
        "bus_io_dev_csm",
        "keypad_csm",
        "wgnd_csm",
        "sounder_csm",
        "repeater_csm",
        "error_code",
    ),
    "area_get_configured": ("areas", "block_count"),
    "area_get_attribs": (
        "area_id",
        "name",
        "flags",
        "kp_single_key_arm",
        "entry1_time",
        "exit_time",
        "ext_entry_time",
        "aa_warn_mins",
        "auto_arm_time",
        "auto_arm_level",
        "sounder",
        "error_code",
    ),
    "area_get_status": (
        "area_id",
        "arm_state",
        "ready_status",
        "alarm_zone",
        "alarms",
        "alarm_mem",
        "zones_bypassed",
        "arm_cmd_state",
        "troubles",
        "chime_count",
        "Chime",
        "xzn_id",
        "xzn_time",
        "ee_timer",
        "alrm_snd",
        "auto_arm_timer",
        "error_code",
    ),
    "area_set_status": ("area_id", "error_code"),
    "area_set_arm_state": ("area_id", "error_code"),
    "zone_get_configured": ("zones", "block_count"),
    "zone_get_attribs": ("zone_id", "name", "area_id", "definition", "flags", "error_code"),
    "zone_get_status": ("zone_id", "secure_state", "contact_state", "BYPASSED", "error_code"),
    "zone_set_status": ("zone_id", "error_code"),
    "zone_get_all_zones_status": ("status", "error_code"),
    "zone_get_defs": ("definitions", "block_count"),
    "zone_get_def_flags": ("definition", "flags", "error_code"),
    "output_get_configured": ("outputs", "block_count"),
    "output_get_attribs": ("output_id", "name", "source_id", "device_id", "hide", "error_code"),
    "output_get_status": ("output_id", "status", "error_code"),
    "output_set_status": ("output_id", "error_code"),
    "output_get_all_outputs_status": ("status", "block_count"),
    "output_get_available": ("outputs", "error_code"),
    "light_get_configured": ("lights", "block_count"),
    "light_get_attribs": ("light_id", "error_code"),
    "light_get_status": ("light_id", "error_code"),
    "light_set_status": ("light_id", "error_code"),
    "light_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "barrier_get_configured": ("barriers", "block_count"),
    "barrier_get_attribs": ("barrier_id", "error_code"),
    "barrier_get_status": ("barrier_id", "error_code"),
    "barrier_set_status": ("barrier_id", "error_code"),
    "barrier_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "lock_get_configured": ("locks", "block_count"),
    "lock_get_attribs": ("lock_id", "error_code"),
    "lock_get_status": ("lock_id", "error_code"),
    "lock_set_status": ("lock_id", "error_code"),
    "lock_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "tstat_get_configured": ("tstats", "block_count"),
    "tstat_get_attribs": ("tstat_id", "error_code"),
    "tstat_set_status": ("tstat_id", "error_code"),
    "rule_get_rules": ("rules", "block_count"),
    "system_get_trouble": ("troubles", "at", "error_code"),
    "log_get_index": ("newest", "max", "total", "error_code"),
    "log_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "log_get_trouble": ("log_full", "error_code"),
    "log_get_attribs": ("log_flags", "sdlog_flags", "error_code"),
    "log_get_list": ("id", "time", "text", "error_code"),
    "log_get_log": ("log_id", "log_count", "error_code"),
    "user_get_configured": ("users", "block_count"),
    "user_get_attribs": ("user_id", "name", "pin", "group_id", "flags", "error_code"),
    "keypad_get_configured": ("keypads", "block_count"),
    "keypad_get_attribs": (
        "keypad_id",
        "source_id",
        "device_id",
        "name",
        "area",
        "backlight",
        "zone_id",
        "quiet_start",
        "quiet_end",
        "flags",
        "error_code",
    ),
    "area_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "zone_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "output_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "tstat_get_table_info": (
        "table_csm",
        "table_elements",
        "tablesize",
        "increment_size",
        "element_size",
        "error_code",
    ),
    "tstat_get_status": (
        "tstat_id",
        "temperature",
        "mode",
        "fan_mode",
        "heat_setpoint",
        "cool_setpoint",
        "humidity",
        "rssi",
        "prec",
        "battery level",
        "error_code",
    ),
    "network_param_get_ssid": ("value", "error_code"),
    "control_get_version_info": (
        "sn",
        "hw",
        "boot",
        "app",
        "db1",
        "db1ver",
        "schema",
        "SSP",
        "BUILD_DATETIME",
        "error_code",
    ),
}

_ALTERNATE_REQUIRED: dict[str, tuple[tuple[str, ...], ...]] = {
    "area_get_configured": (("areas", "block_id", "block_count", "error_code"),),
    "zone_get_configured": (("zones", "block_id", "block_count", "error_code"),),
    "zone_get_defs": (("definitions", "block_id", "block_count", "error_code"),),
    "output_get_configured": (("outputs", "block_id", "block_count", "error_code"),),
    "output_get_all_outputs_status": (("status", "error_code"),),
    "output_get_available": (
        ("outputs", "error_code"),
        ("output_ids", "error_code"),
        ("configured_outputs", "error_code"),
        ("configured_output_ids", "error_code"),
    ),
    "light_get_configured": (("lights", "block_id", "block_count", "error_code"),),
    "barrier_get_configured": (("barriers", "block_id", "block_count", "error_code"),),
    "lock_get_configured": (("locks", "block_id", "block_count", "error_code"),),
    "tstat_get_configured": (("tstats", "block_id", "block_count", "error_code"),),
    "rule_get_rules": (("data", "block_id", "block_count", "error_code"),),
    "user_get_configured": (("users", "block_id", "block_count", "error_code"),),
    "keypad_get_configured": (("keypads", "block_id", "block_count", "error_code"),),
    "log_get_trouble": (("troubles", "error_code"),),
}

_OPTIONAL_KEYS: dict[str, tuple[str, ...]] = {
    "zone_get_status": ("rssi", "low_batt"),
}


def assert_payload_shape(command_key: str, payload: Mapping[str, object] | None) -> None:
    if payload is None:
        raise AssertionError(f"{command_key}: missing payload")

    required = _REQUIRED_KEYS.get(command_key)
    if required is None:
        raise AssertionError(f"{command_key}: no payload shape registered")

    target = _select_payload(command_key, payload)
    optional = _OPTIONAL_KEYS.get(command_key, ())
    required_sets = [required]
    required_sets.extend(_ALTERNATE_REQUIRED.get(command_key, ()))
    missing: list[str] = []
    for required_set in required_sets:
        missing = [key for key in required_set if key not in target and key not in optional]
        if not missing:
            return
    raise AssertionError(f"{command_key}: missing keys {missing} in payload {target!r}")


def _select_payload(command_key: str, payload: Mapping[str, object]) -> Mapping[str, object]:
    if command_key == "system_get_trouble":
        nested = payload.get("get_troubles")
        if isinstance(nested, Mapping):
            return cast(Mapping[str, object], nested)
    return payload


def expected_keys(command_key: str) -> Iterable[str]:
    return _REQUIRED_KEYS.get(command_key, ())



from __future__ import annotations

import logging
import os
from dataclasses import dataclass

import pytest

from elke27_lib import linking
from elke27_lib.client import Elke27Client
from elke27_lib.errors import AuthorizationRequired, Elke27PinRequiredError
from test.helpers.payload_validation import assert_payload_shape


def _env(name: str, default: str | None = None) -> str | None:
    value = os.environ.get(name, default)
    if value == "":
        return default
    return value


def _require_env(name: str) -> str:
    value = _env(name)
    if not value:
        pytest.skip(f"Missing {name} for live network tests.")
    return value


@dataclass(frozen=True, slots=True)
class _Credentials:
    access_code: str
    passphrase: str


@pytest.mark.integration
@pytest.mark.asyncio
async def test_live_network_param_getters() -> None:
    log_level = str(_env("LOG_LEVEL", "INFO") or "INFO").upper()
    logging.basicConfig(level=getattr(logging, log_level, logging.INFO), force=True)
    logging.getLogger("elke27_lib.session").setLevel(logging.DEBUG)

    host = _require_env("ELKE27_HOST")
    port = int(_env("ELKE27_PORT", "2101") or 2101)
    access_code = _require_env("ELKE27_ACCESS_CODE")
    passphrase = _require_env("ELKE27_PASSPHRASE")
    mn = _require_env("ELKE27_MN")
    sn = _require_env("ELKE27_SN")
    fwver = _require_env("ELKE27_FWVER")
    hwver = _require_env("ELKE27_HWVER")
    osver = _require_env("ELKE27_OSVER")
    timeout_s = float(_env("ELKE27_TIMEOUT_S", "10.0") or 10.0)

    client_identity = linking.E27Identity(mn=mn, sn=sn, fwver=fwver, hwver=hwver, osver=osver)
    client = Elke27Client(features=["elke27_lib.features.network_param"])

    creds = _Credentials(access_code=access_code, passphrase=passphrase)
    try:
        link_keys = await client.async_link(
            host,
            port,
            access_code=creds.access_code,
            passphrase=creds.passphrase,
            client_identity=client_identity,
            timeout_s=timeout_s,
        )
        await client.async_connect(host, port, link_keys)
    except Exception as exc:  # noqa: BLE001
        pytest.fail(f"Link/connect failed: {exc}")

    pin_value = _env("ELKE27_PIN")
    if not pin_value:
        await client.async_disconnect()
        pytest.skip("ELKE27_PIN not set; network_param_get_ssid requires installer PIN.")

    result = await client.async_execute(
        "network_param_get_ssid",
        pin=pin_value,
        timeout_s=timeout_s,
    )

    if not result.ok:
        if isinstance(result.error, AuthorizationRequired):
            print("Authorization required for SSID.")
        elif isinstance(result.error, Elke27PinRequiredError):
            pytest.skip(f"network_param_get_ssid auth not available: {result.error}")
        else:
            await client.async_disconnect()
            pytest.fail(f"network_param_get_ssid failed: {result.error}")
    else:
        assert_payload_shape("network_param_get_ssid", result.data)
        print("network_param_get_ssid succeeded.")

    await client.async_disconnect()



from __future__ import annotations

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import Event
from elke27_lib.generators.bus_ios import generator_bus_ios_get_trouble
from elke27_lib.generators.control import generator_control_get_trouble
from elke27_lib.generators.log import generator_log_get_trouble
from elke27_lib.handlers.bus_ios import make_bus_ios_get_trouble_handler
from elke27_lib.handlers.control import make_control_get_trouble_handler
from elke27_lib.handlers.log import make_log_get_trouble_handler
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[Event] = []

    def __call__(self, evt: Event, _ctx: DispatchContext) -> None:
        self.events.append(evt)


_Ctx = make_ctx


def test_generator_control_get_trouble() -> None:
    payload, response_key = generator_control_get_trouble()
    assert payload == {}
    assert response_key == ("control", "get_trouble")


def test_generator_log_get_trouble() -> None:
    payload, response_key = generator_log_get_trouble()
    assert payload == {}
    assert response_key == ("log", "get_trouble")


def test_generator_bus_ios_get_trouble() -> None:
    payload, response_key = generator_bus_ios_get_trouble()
    assert payload == {}
    assert response_key == ("bus_io_dev", "get_trouble")


def test_control_get_trouble_handler_updates_state() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_control_get_trouble_handler(state, emit, now=lambda: 123.0)

    msg = {"control": {"get_trouble": {"modtbl": ["system"], "error_code": 0}}}
    assert handler(msg, _Ctx()) is True
    status = state.control_status.get("get_trouble")
    assert isinstance(status, dict)
    assert status["modtbl"] == ["system"]


def test_log_get_trouble_handler_updates_state() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_log_get_trouble_handler(state, emit, now=lambda: 123.0)

    msg = {"log": {"get_trouble": {"log_full": True, "error_code": 0}}}
    assert handler(msg, _Ctx()) is True
    status = state.log_status.get("get_trouble")
    assert isinstance(status, dict)
    assert status["log_full"] is True


def test_bus_ios_get_trouble_handler_updates_state() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_bus_ios_get_trouble_handler(state, emit, now=lambda: 123.0)

    msg = {"bus_io_dev": {"get_trouble": {"MISSING": ["11000001"], "error_code": 0}}}
    assert handler(msg, _Ctx()) is True
    status = state.bus_io_status.get("get_trouble")
    assert isinstance(status, dict)
    assert status["MISSING"] == ["11000001"]



from __future__ import annotations

import importlib

import pytest

from elke27_lib import encryption as enc


def test_hex_to_bytes_valid() -> None:
    assert enc.hex_to_bytes("00ff") == bytes([0x00, 0xFF])


@pytest.mark.parametrize("value", [None, "", " ", "0", "abc", "zz"])
def test_hex_to_bytes_invalid(value: str | None) -> None:
    with pytest.raises(enc.E27CryptoError):
        enc.hex_to_bytes(value)  # type: ignore[arg-type]


def test_swap_endianness_basic() -> None:
    assert enc.swap_endianness(b"\x01\x02\x03\x04") == b"\x04\x03\x02\x01"
    assert (
        enc.swap_endianness(b"\x01\x02\x03\x04\x10\x11\x12\x13")
        == b"\x04\x03\x02\x01\x13\x12\x11\x10"
    )


@pytest.mark.parametrize("value", [None, b"", b"\x01", b"\x01\x02"])
def test_swap_endianness_invalid(value: bytes | None) -> None:
    with pytest.raises(enc.E27CryptoError):
        enc.swap_endianness(value)  # type: ignore[arg-type]


def test_calculate_block_padding() -> None:
    assert enc.calculate_block_padding(0) == 0
    assert enc.calculate_block_padding(1) == 15
    assert enc.calculate_block_padding(16) == 0


def test_calculate_block_padding_negative() -> None:
    with pytest.raises(enc.E27CryptoError):
        enc.calculate_block_padding(-1)


def test_schema0_encrypt_decrypt_roundtrip() -> None:
    key = b"\x00" * 16
    pt = b"\x01" * 16
    ct = enc.encrypt_schema0_plaintext(key=key, plaintext=pt)
    assert ct != pt
    out = enc.decrypt_schema0_ciphertext(key=key, ciphertext=ct)
    assert out == pt


def test_schema0_invalid_lengths() -> None:
    key = b"\x00" * 16
    with pytest.raises(enc.E27CryptoError):
        enc.encrypt_schema0_plaintext(key=key, plaintext=b"")
    with pytest.raises(enc.E27CryptoError):
        enc.decrypt_schema0_ciphertext(key=key, ciphertext=b"")


def test_key_conversions() -> None:
    raw = bytes(range(16)).hex()
    assert enc.tempkey_hex_to_aes_key(raw) == enc.swap_endianness(bytes(range(16)))
    assert enc.linkkey_hex_to_aes_key(raw) == enc.swap_endianness(bytes(range(16)))
    assert enc.sessionkey_hex_to_aes_key(raw) == bytes(range(16))


def test_key_conversions_invalid_length() -> None:
    with pytest.raises(enc.E27CryptoError):
        enc.tempkey_hex_to_aes_key("00")
    with pytest.raises(enc.E27CryptoError):
        enc.linkkey_hex_to_aes_key("00")
    with pytest.raises(enc.E27CryptoError):
        enc.sessionkey_hex_to_aes_key("00")


def test_decrypt_hello_field_roundtrip() -> None:
    key_hex = "00112233445566778899aabbccddeeff"
    key = enc.linkkey_hex_to_aes_key(key_hex)
    pt = b"\x10" * 16
    ct = enc._aes_cbc_encrypt_no_padding(key, enc.API_LINK_IV, enc.swap_endianness(pt))
    ct_hex = enc.swap_endianness(ct).hex()
    out = enc.decrypt_hello_field(linkkey_hex=key_hex, field_hex=ct_hex)
    assert out == pt


def test_crypto_backend_missing(monkeypatch: pytest.MonkeyPatch) -> None:
    module = importlib.import_module("elke27_lib.encryption")
    monkeypatch.setattr(module, "_has_crypto", False)
    monkeypatch.setattr(module, "Cipher", None)
    monkeypatch.setattr(module, "algorithms", None)
    monkeypatch.setattr(module, "modes", None)
    monkeypatch.setattr(module, "default_backend", None)
    with pytest.raises(enc.E27CryptoError):
        module._aes_cbc_encrypt_no_padding(b"\x00" * 16, enc.API_LINK_IV, b"\x00" * 16)
    with pytest.raises(enc.E27CryptoError):
        module._aes_cbc_decrypt_no_padding(b"\x00" * 16, enc.API_LINK_IV, b"\x00" * 16)


def test_require_helpers_errors() -> None:
    with pytest.raises(enc.E27CryptoError):
        enc._require_block_multiple(None, 16, "data")  # type: ignore[arg-type]
    with pytest.raises(enc.E27CryptoError):
        enc._require_block_multiple(b"", 16, "data")
    with pytest.raises(enc.E27CryptoError):
        enc._require_block_multiple(b"\x00" * 3, 4, "data")
    with pytest.raises(enc.E27CryptoError):
        enc._require_len(None, 16, "key")  # type: ignore[arg-type]


def test_import_without_crypto(monkeypatch: pytest.MonkeyPatch) -> None:
    import builtins

    real_import = builtins.__import__

    def _fake_import(name, globals=None, locals=None, fromlist=(), level=0):  # type: ignore[no-untyped-def]
        if name.startswith("cryptography"):
            raise ImportError("no crypto")
        return real_import(name, globals, locals, fromlist, level)

    with monkeypatch.context() as m:
        m.setattr(builtins, "__import__", _fake_import)
        module = importlib.reload(enc)
        assert module._has_crypto is False
        assert module.Cipher is None
        assert module.algorithms is None
        assert module.modes is None
        assert module.default_backend is None

    importlib.reload(enc)



"""
E27 JSON Dispatcher (Skeleton)

Key points (DDR-0036 aligned):
- Input: msg is a JSON object (dict) emitted by Session.recv_json()/pump_once()/on_message.
- Route extraction is deterministic:
    - domain = the only non-meta root key (meta keys: seq, session_id)
    - name   = if msg[domain] is a dict with exactly one key -> that key
               else fall back to "__root__"/"__empty__"/"__value__" with DispatchError(s)
  NOTE: The inner value under the command key may be a multi-key dict of parameters. That is expected.
- Broadcast vs directed classification is based ONLY on root seq:
    - seq == 0 => BROADCAST
    - seq > 0  => DIRECTED
    - missing  => UNKNOWN (no error; hello/bootstrap often lack root seq)
    - invalid/negative => UNKNOWN + ERR_INVALID_SEQ
- Seq correlation (optional) is seq-first and route-second. Dispatcher does not assume response route == request route.
- No nested-seq heuristics. Hello/bootstrap messages remain routable but not correlatable unless root seq exists.
- Dispatcher never raises for unknown/ambiguous JSON shapes; it only raises for programmer error (non-mapping input).
- Reserved error channel: ("__error__", "<error_code>") and ("__error__", "__all__")
  Error handlers receive an *error envelope message* (not the original msg) so they don't have to re-derive details.
"""

from __future__ import annotations

import json
import logging
import time
from collections.abc import Callable, Iterable, Mapping
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, cast

RouteKey = tuple[str, str]


class MessageKind(str, Enum):
    UNKNOWN = "UNKNOWN"
    DIRECTED = "DIRECTED"
    BROADCAST = "BROADCAST"


class DispatchSeverity(str, Enum):
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


@dataclass(frozen=True)
class DispatchError:
    code: str
    message: str
    domain: str | None = None
    name: str | None = None
    keys: tuple[str, ...] = ()
    severity: DispatchSeverity = DispatchSeverity.WARNING


@dataclass(frozen=True)
class PendingRequest:
    """
    Minimal pending request record. Seq correlation is seq-first and route-second.
    Policy (timeouts/retries/backoff) belongs above Dispatcher.
    """

    seq: int
    expected_route: RouteKey | None = None
    created_at: float | None = None
    opaque: Any = None


@dataclass(frozen=True)
class PagedTransferKey:
    """
    Correlation key for paged transfers (ADR-0013).
    """

    session_id: int | None
    transfer_id: int
    route: RouteKey


@dataclass(frozen=True)
class PagedBlock:
    block_id: int
    payload: Mapping[str, Any]


PagedMergeFn = Callable[[list[PagedBlock], int], Mapping[str, Any]]
PagedRequestBlockFn = Callable[[int, PagedTransferKey], None]


@dataclass(frozen=True)
class PagedRouteSpec:
    merge_fn: PagedMergeFn
    request_block: PagedRequestBlockFn | None
    timeout_s: float


@dataclass
class PagedTransfer:
    key: PagedTransferKey
    total_count: int | None
    created_at: float
    last_update_at: float
    received_blocks: dict[int, Mapping[str, Any]] = field(default_factory=dict)
    requested_blocks: set[int] = field(default_factory=set)


@dataclass(frozen=True)
class DispatchContext:
    """
    Diagnostics-only context. No policy.
    """

    kind: MessageKind
    seq: int | None
    session_id: int | None
    route: RouteKey
    classification: str  # "BROADCAST" | "RESPONSE" | "UNSOLICITED" | "UNKNOWN"
    response_match: PendingRequest | None = None
    raw_route: RouteKey | None = (
        None  # original route before any error routing (usually same as route)
    )


@dataclass
class DispatchResult:
    route: RouteKey
    kind: MessageKind
    seq: int | None
    session_id: int | None
    classification: str
    response_match: PendingRequest | None = None
    errors: list[DispatchError] = field(default_factory=list)
    handled: bool = False


# Handler signature:
# Return True means "this handler recognized/handled it" (fan-out still continues).
DispatchHandler = Callable[[Mapping[str, Any], DispatchContext], bool]


META_KEYS = ("seq", "session_id")
ERROR_DOMAIN = "__error__"
ERROR_ALL = "__all__"
ERROR_ROOT = "panel_error"

LOG = logging.getLogger(__name__)


# Stable error codes (dispatcher-generated; never extracted from JSON)
ERR_ROOT_EMPTY = "root_empty"
ERR_ROOT_MULTI = "root_multi"
ERR_DOMAIN_EMPTY = "domain_empty"
ERR_DOMAIN_MULTI = "domain_multi"
ERR_UNEXPECTED_VALUE_TYPE = "unexpected_value_type"
ERR_INVALID_SEQ = "invalid_seq"


def _payload_preview(msg: Mapping[str, Any], *, limit: int = 512) -> str:
    try:
        text = json.dumps(msg, sort_keys=True, separators=(",", ":"))
    except (TypeError, ValueError):
        text = repr(msg)
    if len(text) > limit:
        return f"{text[:limit]}..."
    return text


class Dispatcher:
    """
    Deterministic route extraction + handler fan-out + seq correlation.

    This skeleton intentionally does not implement domain semantics. It only:
    - extracts route
    - classifies broadcast/directed
    - optionally correlates by seq
    - calls registered handlers
    - emits DispatchError notifications
    """

    def __init__(
        self,
        *,
        now: Callable[[], float] = time.monotonic,
        paged_timeout_s: float = 10.0,
    ) -> None:
        self._handlers: dict[RouteKey, list[DispatchHandler]] = {}
        self._pending: dict[int, PendingRequest] = {}
        self._paged_routes: dict[RouteKey, PagedRouteSpec] = {}
        self._paged_transfers: dict[PagedTransferKey, PagedTransfer] = {}
        self._now: Callable[[], float] = now
        self._paged_timeout_s: float = paged_timeout_s

    # --- Registration API ---

    def register(self, route: RouteKey, handler: DispatchHandler) -> None:
        """
        Register a handler for a route.

        Reserved:
        - ("__error__", "<error_code>") for dispatch errors
        - ("__error__", "__all__") catch-all errors
        """
        self._handlers.setdefault(route, []).append(handler)

    def register_domain(self, domain: str, handler: DispatchHandler) -> None:
        """
        Convenience: register a domain-level handler for (domain, "__root__").
        This is where ambiguous domain payloads are routed.
        """
        self.register((domain, "__root__"), handler)

    def register_paged(
        self,
        route: RouteKey,
        *,
        merge_fn: PagedMergeFn,
        request_block: PagedRequestBlockFn | None = None,
        timeout_s: float | None = None,
    ) -> None:
        """
        Register a paged route for ADR-0013 reassembly.

        Handlers for paged routes receive only fully assembled payloads.
        """
        self._paged_routes[route] = PagedRouteSpec(
            merge_fn=merge_fn,
            request_block=request_block,
            timeout_s=timeout_s if timeout_s is not None else self._paged_timeout_s,
        )

    def is_paged(self, route: RouteKey) -> bool:
        return route in self._paged_routes

    def unregister(self, route: RouteKey, handler: DispatchHandler) -> None:
        handlers = self._handlers.get(route)
        if not handlers:
            return
        try:
            handlers.remove(handler)
        except ValueError:
            return
        if not handlers:
            self._handlers.pop(route, None)

    # --- Pending request API (optional) ---

    def add_pending(self, pending: PendingRequest) -> None:
        self._pending[pending.seq] = pending

    def match_pending(self, seq: int, *, pop: bool = True) -> PendingRequest | None:
        """
        Seq correlation is seq-first. The default policy is one-shot (pop=True).
        """
        if pop:
            return self._pending.pop(seq, None)
        return self._pending.get(seq)

    def drop_pending(self, seq: int) -> None:
        self._pending.pop(seq, None)

    def pending_count(self) -> int:
        return len(self._pending)

    # --- Dispatch API ---

    def dispatch(self, msg: Mapping[str, Any]) -> DispatchResult:
        """
        Dispatch a single inbound message dict.

        Never raises for unknown/ambiguous routing; returns errors in DispatchResult and emits __error__ handlers.
        Raises TypeError only for programmer error (non-mapping input).
        """
        route, route_errors = self._extract_route(msg)
        seq, kind, seq_errors = self._classify_kind(msg)

        errors: list[DispatchError] = []
        errors.extend(route_errors)
        errors.extend(seq_errors)

        # Correlation / classification
        classification = "UNKNOWN"
        response_match: PendingRequest | None = None
        if kind == MessageKind.BROADCAST:
            classification = "BROADCAST"
        elif kind == MessageKind.DIRECTED and seq is not None:
            pending = self.match_pending(seq, pop=True)
            if pending is not None:
                classification = "RESPONSE"
                response_match = pending
            else:
                classification = "UNSOLICITED"
        else:
            classification = "UNKNOWN"

        session_id = self._get_int(msg, "session_id")

        ctx = DispatchContext(
            kind=kind,
            seq=seq,
            session_id=session_id,
            route=route,
            classification=classification,
            response_match=response_match,
            raw_route=route,
        )

        result = DispatchResult(
            route=route,
            kind=kind,
            seq=seq,
            session_id=session_id,
            classification=classification,
            response_match=response_match,
            errors=errors,
            handled=False,
        )

        # Abort stale transfers before handling new traffic.
        self._expire_paged_transfers()

        # ADR-0013 paged reassembly (short-circuit partial blocks).
        paged_msg = self._maybe_reassemble_paged(msg, ctx, response_match)
        if paged_msg is None:
            # Partial block consumed; no handler dispatch.
            if errors:
                self._emit_errors(ctx, errors, msg=msg)
            result.handled = True
            return result
        msg = paged_msg

        # Normal dispatch fan-out (assembled messages only for paged routes)
        if LOG.isEnabledFor(logging.DEBUG) and ctx.route[0] == "zone":
            LOG.debug(
                "dispatch zone message: route=%s classification=%s seq=%s",
                ctx.route,
                ctx.classification,
                ctx.seq,
            )
        result.handled = self._dispatch_normal(msg, ctx)

        # Emit dispatch errors (after normal routing)
        if errors:
            self._emit_errors(ctx, errors, msg=msg)

        return result

    def _maybe_reassemble_paged(
        self,
        msg: Mapping[str, Any],
        ctx: DispatchContext,
        response_match: PendingRequest | None,
    ) -> Mapping[str, Any] | None:
        route = ctx.route
        if route not in self._paged_routes:
            return msg

        transfer_key = self._paged_transfer_key(route, ctx.session_id, response_match)
        if transfer_key is None:
            return msg

        payload = self._extract_paged_payload(msg, route)
        if payload is None:
            return msg

        # Domain-root auth error aborts reassembly and lets handlers surface it.
        domain_obj = msg.get(route[0])
        if isinstance(domain_obj, Mapping):
            domain_map = cast(Mapping[str, Any], domain_obj)
            error_code = domain_map.get("error_code")
            if isinstance(error_code, int) and error_code == 11008:
                self._paged_transfers.pop(transfer_key, None)
                return msg

        payload_error = payload.get("error_code")
        if isinstance(payload_error, int) and payload_error == 11008:
            self._paged_transfers.pop(transfer_key, None)
            return msg

        block_id = payload.get("block_id")
        block_count = payload.get("block_count")
        if not isinstance(block_id, int) or block_id < 1:
            LOG.warning("paged block missing/invalid block_id for route=%s", route)
            return None
        if not isinstance(block_count, int) or block_count < 1:
            LOG.warning("paged block missing/invalid block_count for route=%s", route)
            return None
        if block_id > block_count:
            LOG.warning(
                "paged block_id out of range for route=%s block_id=%s block_count=%s",
                route,
                block_id,
                block_count,
            )
            return None

        transfer = self._paged_transfers.get(transfer_key)
        now = self._now()
        if transfer is None:
            transfer = PagedTransfer(
                key=transfer_key,
                total_count=block_count,
                created_at=now,
                last_update_at=now,
            )
            self._paged_transfers[transfer_key] = transfer
        elif transfer.total_count is not None and transfer.total_count != block_count:
            self._paged_transfers.pop(transfer_key, None)
            return msg

        if block_id in transfer.received_blocks:
            return None

        transfer.received_blocks[block_id] = payload
        transfer.last_update_at = now
        transfer.requested_blocks.add(block_id)

        spec = self._paged_routes[route]
        if transfer.total_count is None:
            transfer.total_count = block_count

        # Request the next missing block, if any, using the transfer key.
        if spec.request_block is not None and transfer.total_count:
            for next_block in range(1, transfer.total_count + 1):
                if next_block in transfer.received_blocks:
                    continue
                if next_block in transfer.requested_blocks:
                    continue
                try:
                    spec.request_block(next_block, transfer_key)
                except Exception as exc:
                    LOG.warning(
                        "Paged request_block failed: route=%s.%s block=%s key=%s error=%s",
                        route[0],
                        route[1],
                        next_block,
                        transfer_key,
                        exc,
                        exc_info=True,
                    )
                    break
                transfer.requested_blocks.add(next_block)
                break

        if transfer.total_count and len(transfer.received_blocks) < transfer.total_count:
            return None

        blocks = [
            PagedBlock(block_id=bid, payload=transfer.received_blocks[bid])
            for bid in sorted(transfer.received_blocks.keys())
        ]
        merged_payload = spec.merge_fn(blocks, transfer.total_count or len(blocks))
        assembled = dict(msg)
        assembled[route[0]] = {route[1]: dict(merged_payload)}
        self._paged_transfers.pop(transfer_key, None)
        return assembled

    def _paged_transfer_key(
        self,
        route: RouteKey,
        session_id: int | None,
        response_match: PendingRequest | None,
    ) -> PagedTransferKey | None:
        if response_match is not None and isinstance(response_match.opaque, PagedTransferKey):
            return response_match.opaque
        if response_match is not None:
            return PagedTransferKey(
                session_id=session_id, transfer_id=response_match.seq, route=route
            )
        return None

    def _extract_paged_payload(
        self, msg: Mapping[str, Any], route: RouteKey
    ) -> Mapping[str, Any] | None:
        domain, name = route
        domain_obj = msg.get(domain)
        if not isinstance(domain_obj, Mapping):
            return None
        domain_map = cast(Mapping[str, Any], domain_obj)
        payload = domain_map.get(name)
        if not isinstance(payload, Mapping):
            return None
        payload_map = cast(Mapping[str, Any], payload)
        return payload_map

    def _expire_paged_transfers(self) -> None:
        if not self._paged_transfers:
            return
        now = self._now()
        expired: list[PagedTransferKey] = []
        for key, transfer in self._paged_transfers.items():
            spec = self._paged_routes.get(key.route)
            timeout_s = spec.timeout_s if spec is not None else self._paged_timeout_s
            if now - transfer.last_update_at >= timeout_s:
                expired.append(key)
        for key in expired:
            self._paged_transfers.pop(key, None)

    def abort_paged_transfers(self) -> None:
        """
        Abort all in-progress paged transfers (e.g., on disconnect).
        """
        self._paged_transfers.clear()

    # --- Internal: routing + classification ---

    def _extract_route(self, msg: Mapping[str, Any]) -> tuple[RouteKey, list[DispatchError]]:
        errors: list[DispatchError] = []
        root_non_meta = [k for k in msg if k not in META_KEYS]

        if self._is_root_error_envelope(msg):
            return (ERROR_DOMAIN, ERROR_ROOT), errors

        if len(root_non_meta) == 0:
            errors.append(
                DispatchError(
                    code=ERR_ROOT_EMPTY,
                    message="No domain keys present at root (only meta keys found).",
                    domain="__root__",
                    name="__empty__",
                    keys=tuple(msg.keys()),
                )
            )
            return ("__root__", "__empty__"), errors

        if len(root_non_meta) > 1:
            errors.append(
                DispatchError(
                    code=ERR_ROOT_MULTI,
                    message="Multiple domain keys present at root; cannot determine a single domain.",
                    domain="__root__",
                    name="__multi__",
                    keys=tuple(msg.keys()),  # include meta keys too for full diagnostics
                )
            )
            return ("__root__", "__multi__"), errors

        domain = root_non_meta[0]
        v = msg.get(domain)

        if isinstance(v, Mapping):
            v_map = cast(Mapping[str, Any], v)
            if self._is_domain_root_error(v_map):
                return (domain, "error"), errors

            inner_keys = list(v_map.keys())
            if len(inner_keys) == 0:
                errors.append(
                    DispatchError(
                        code=ERR_DOMAIN_EMPTY,
                        message=f"Domain '{domain}' object is empty; cannot determine message name.",
                        domain=domain,
                        name="__empty__",
                    )
                )
                return (domain, "__empty__"), errors

            if len(inner_keys) == 1:
                # Contract: name is the sole key under the domain object.
                return (domain, inner_keys[0]), errors

            # Multi-key domain object: treat as domain-level payload.
            errors.append(
                DispatchError(
                    code=ERR_DOMAIN_MULTI,
                    message=f"Domain '{domain}' object contains multiple keys; routing to domain-level handler.",
                    domain=domain,
                    name="__root__",
                    keys=tuple(inner_keys),
                )
            )
            return (domain, "__root__"), errors

        errors.append(
            DispatchError(
                code=ERR_UNEXPECTED_VALUE_TYPE,
                message=(
                    f"Domain '{domain}' value is unexpected type '{type(v).__name__}'; "
                    "routing to domain-level handler."
                ),
                domain=domain,
                name="__value__",
            )
        )
        return (domain, "__value__"), errors

    @staticmethod
    def _is_domain_root_error(domain_value: Mapping[str, Any]) -> bool:
        error_code = domain_value.get("error_code")
        if isinstance(error_code, str):
            try:
                error_code = int(error_code)
            except ValueError:
                error_code = None
        if not isinstance(error_code, int):
            return False

        # If there's an obvious command key (non-error key whose value is a mapping), treat it as normal routing.
        for key, value in domain_value.items():
            if key in ("error_code", "error_message", "error_text"):
                continue
            if isinstance(value, Mapping):
                return False

        return True

    @staticmethod
    def _is_root_error_envelope(msg: Mapping[str, Any]) -> bool:
        error_code = msg.get("error_code")
        if isinstance(error_code, str):
            try:
                error_code = int(error_code)
            except ValueError:
                error_code = None
        if not isinstance(error_code, int):
            return False

        error_message = msg.get("error_message") or msg.get("error_text")
        if not isinstance(error_message, str) or not error_message:
            return False

        allowed_keys = {
            "error_code",
            "error_message",
            "error_text",
            "error_detail",
            "error_payload",
            "payload",
            "detail",
        }
        root_non_meta = [k for k in msg if k not in META_KEYS]
        return all(k in allowed_keys for k in root_non_meta)

    def _classify_kind(
        self, msg: Mapping[str, Any]
    ) -> tuple[int | None, MessageKind, list[DispatchError]]:
        errors: list[DispatchError] = []

        # Missing seq is allowed and yields UNKNOWN without error (bootstrap/hello often lack root seq).
        if "seq" not in msg:
            return None, MessageKind.UNKNOWN, errors

        seq_val = msg.get("seq")
        if not isinstance(seq_val, int):
            errors.append(
                DispatchError(
                    code=ERR_INVALID_SEQ,
                    message=f"Invalid seq type '{type(seq_val).__name__}' (expected int).",
                    keys=("seq",),
                )
            )
            return None, MessageKind.UNKNOWN, errors

        if seq_val == 0:
            return 0, MessageKind.BROADCAST, errors
        if seq_val > 0:
            return seq_val, MessageKind.DIRECTED, errors

        # Negative or otherwise invalid values: normalize seq to None to simplify downstream.
        errors.append(
            DispatchError(
                code=ERR_INVALID_SEQ,
                message="Invalid seq value (expected 0 for broadcast or >0 for directed).",
                keys=("seq",),
            )
        )
        return None, MessageKind.UNKNOWN, errors

    @staticmethod
    def _get_int(msg: Mapping[str, Any], key: str) -> int | None:
        v = msg.get(key)
        return v if isinstance(v, int) else None

    # --- Internal: handler fan-out ---

    def _dispatch_normal(self, msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        """
        Invoke handlers for:
        - exact route (domain,name)
        - domain-level fallback (domain,"__root__") for ambiguous cases
        Fan-out policy: call all handlers; return indicates "handled" but does not stop others.
        """
        domain, name = ctx.route
        handled = False

        # Exact route
        handled |= self._call_handlers((domain, name), msg, ctx)

        # Domain-level fallback for ambiguous routing; avoid double-dispatch when exact is already __root__.
        if name in ("__empty__", "__value__"):
            handled |= self._call_handlers((domain, "__root__"), msg, ctx)

        if name == "__root__":
            # __root__ is itself the domain-level route; no additional call needed.
            pass

        # Optional: root diagnostics handlers for "__root__" errors
        if domain == "__root__" and name in ("__empty__", "__multi__"):
            handled |= self._call_handlers(("__root__", "__root__"), msg, ctx)

        return handled

    def _emit_errors(
        self, ctx: DispatchContext, errors: Iterable[DispatchError], *, msg: Mapping[str, Any]
    ) -> None:
        """
        Emit dispatch errors to:
        - ("__error__", error.code)
        - ("__error__", "__all__")

        Error handlers receive an *error envelope message* that contains the DispatchError details.
        """
        for err in errors:
            err_msg: Mapping[str, Any] = {
                "seq": ctx.seq,
                "session_id": ctx.session_id,
                ERROR_DOMAIN: {
                    err.code: {
                        "message": err.message,
                        "domain": err.domain,
                        "name": err.name,
                        "keys": list(err.keys),
                        "severity": err.severity.value,
                        "payload": _payload_preview(msg),
                    }
                },
            }

            err_ctx = DispatchContext(
                kind=ctx.kind,
                seq=ctx.seq,
                session_id=ctx.session_id,
                route=(ERROR_DOMAIN, err.code),
                classification=ctx.classification,
                response_match=ctx.response_match,
                raw_route=ctx.raw_route,
            )

            # Specific code
            self._call_handlers((ERROR_DOMAIN, err.code), err_msg, err_ctx)

            # Catch-all
            self._call_handlers((ERROR_DOMAIN, ERROR_ALL), err_msg, err_ctx)

    def _call_handlers(self, route: RouteKey, msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        handlers = self._handlers.get(route)
        if not handlers:
            return False

        handled_any = False
        for h in list(handlers):
            try:
                handled = bool(h(msg, ctx))
            except (AttributeError, KeyError, RuntimeError, TypeError, ValueError) as exc:
                # Contract: handler exceptions do not break dispatcher.
                LOG.warning(
                    "Handler error: route=%s msg_keys=%s error=%s",
                    route,
                    tuple(msg.keys()),
                    exc,
                    exc_info=True,
                )
                handled = False
            handled_any |= handled
        return handled_any



"""
elke27_lib/handlers/control.py

Read-only handlers for the "control" domain.

Colocation policy:
- Message-specific reconcile helpers live in the same module as their handlers.
- Reconcile helpers are module-private (prefixed with _).
- Reconcile helpers are PURE:
    - no dispatcher context
    - no event emission
    - no I/O / logging
- Handlers:
    - extract payload from msg
    - call reconcile helper(s)
    - emit events

Focus: ("control","get_version_info")
"""

from __future__ import annotations

import logging
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import Any, cast

from elke27_lib.dispatcher import (
    DispatchContext,  # adjust import to your dispatcher module location
)
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    ApiError,
    AuthenticateResult,
    AuthorizationRequiredEvent,
    CsmSnapshotUpdated,
    DispatchRoutingError,
    DomainCsmChanged,
    Event,
    PanelVersionInfoUpdated,
)
from elke27_lib.states import PanelState, update_csm_snapshot

EmitFn = Callable[[Event, DispatchContext], None]
NowFn = Callable[[], float]

LOG = logging.getLogger(__name__)


def _as_mapping(obj: object) -> Mapping[str, Any] | None:
    if isinstance(obj, Mapping):
        return cast(Mapping[str, Any], obj)
    return None


# -------------------------
# Module-private reconcile
# -------------------------


@dataclass(frozen=True, slots=True)
class _VersionInfoOutcome:
    changed_fields: tuple[str, ...]
    error_code: int | None
    warnings: tuple[str, ...]


def _reconcile_control_get_version_info(
    state: PanelState,
    payload: Mapping[str, Any],
    *,
    now: float,
) -> _VersionInfoOutcome:
    """
    Pure reconcile for control.get_version_info.

    v0 rules (conservative):
    - Patch-style only (never clears absent fields)
    - Strict typing for stored panel meta fields (strings only)
    - Always updates state.panel.last_message_at
    """
    warnings: list[str] = []
    changed: set[str] = set()

    # Always update panel freshness
    state.panel.last_message_at = now

    # Optional response field
    error_code = payload.get("error_code")
    if error_code is not None and not isinstance(error_code, int):
        warnings.append(
            f"field 'error_code' wrong type (expected int, got {type(error_code).__name__})"
        )
        error_code = None

    # Canonical stored fields (expand cautiously once live payload confirms keys)
    # Accept a couple common variants to reduce friction for first live test.
    model = _first_present(payload, ("model", "panel_model", "hw", "hwver"))
    firmware = _first_present(payload, ("firmware", "fw", "firmware_version", "sw_version", "SSP"))
    serial = _first_present(payload, ("serial", "serial_number", "sn"))

    if model is not None:
        if isinstance(model, int):
            model = str(model)
        if isinstance(model, str):
            if state.panel.model != model:
                state.panel.model = model
                changed.add("model")
        else:
            warnings.append(
                f"field 'model' wrong type (expected str/int, got {type(model).__name__})"
            )

    if firmware is not None:
        if isinstance(firmware, str):
            if state.panel.firmware != firmware:
                state.panel.firmware = firmware
                changed.add("firmware")
        else:
            warnings.append(
                f"field 'firmware' wrong type (expected str, got {type(firmware).__name__})"
            )

    if serial is not None:
        if isinstance(serial, str):
            if state.panel.serial != serial:
                state.panel.serial = serial
                changed.add("serial")
        else:
            warnings.append(
                f"field 'serial' wrong type (expected str, got {type(serial).__name__})"
            )

    return _VersionInfoOutcome(
        changed_fields=tuple(sorted(changed)),
        error_code=error_code,
        warnings=tuple(warnings),
    )


def _first_present(payload: Mapping[str, Any], keys: tuple[str, ...]) -> Any:
    for k in keys:
        if k in payload:
            return payload.get(k)
    return None


# -------------------------
# Handler factory
# -------------------------


def make_control_authenticate_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("control","authenticate") where payload is msg["control"]["authenticate"].
    """

    def handler_control_authenticate(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        _ = now
        control_obj = _as_mapping(msg.get("control"))
        if control_obj is None:
            return False

        payload = _as_mapping(control_obj.get("authenticate"))
        if payload is None:
            return False

        error_code = payload.get("error_code")
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="control",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
            else:
                emit(
                    ApiError(
                        kind=ApiError.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="control",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
            emit(
                AuthenticateResult(
                    kind=AuthenticateResult.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    success=False,
                    error_code=error_code,
                ),
                ctx,
            )
            _notify_auth_opaque(ctx, success=False, error_code=error_code)
            return True

        emit(
            AuthenticateResult(
                kind=AuthenticateResult.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
                success=True,
                error_code=0,
            ),
            ctx,
        )
        _apply_auth_csm_updates(state, emit, ctx, payload)
        _notify_auth_opaque(ctx, success=True, error_code=0)
        return True

    return handler_control_authenticate


def make_control_get_trouble_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("control","get_trouble") responses.
    """

    def handler_control_get_trouble(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        control_obj = _as_mapping(msg.get("control"))
        if control_obj is None:
            return False

        payload = _as_mapping(control_obj.get("get_trouble"))
        if payload is None:
            return False
        if getattr(ctx, "classification", None) == "BROADCAST":
            LOG.debug("control.get_trouble broadcast received")

        error_code = payload.get("error_code", control_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="control",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="control",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        state.control_status["get_trouble"] = dict(payload)
        state.panel.last_message_at = now()
        return True

    return handler_control_get_trouble


def _notify_auth_opaque(ctx: DispatchContext, *, success: bool, error_code: int | None) -> None:
    match = ctx.response_match
    if match is None:
        return
    opaque = match.opaque
    if opaque is None:
        return
    payload = {"success": success, "error_code": error_code}
    put_nowait = getattr(opaque, "put_nowait", None)
    if callable(put_nowait):
        try:
            put_nowait(payload)
            return
        except Exception as exc:
            LOG.warning("auth opaque put_nowait failed: %s", exc, exc_info=True)
            return
    put = getattr(opaque, "put", None)
    if callable(put):
        try:
            put(payload)
        except Exception as exc:
            LOG.warning("auth opaque put failed: %s", exc, exc_info=True)
            return


def _apply_auth_csm_updates(
    state: PanelState,
    emit: EmitFn,
    ctx: DispatchContext,
    payload: Mapping[str, Any],
) -> None:
    for key, value in payload.items():
        domain = _domain_from_csm_key(key)
        if domain is None:
            continue
        parsed = _coerce_csm_value(key, value, source="authenticate")
        if parsed is None:
            continue
        old = state.domain_csm_by_name.get(domain)
        if old == parsed:
            continue
        state.domain_csm_by_name[domain] = parsed
        emit(
            DomainCsmChanged(
                kind=DomainCsmChanged.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
                csm_domain=domain,
                old=old,
                new=parsed,
            ),
            ctx,
        )

    snapshot = update_csm_snapshot(state)
    if snapshot is not None:
        emit(
            CsmSnapshotUpdated(
                kind=CsmSnapshotUpdated.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
                snapshot=snapshot,
            ),
            ctx,
        )


def _domain_from_csm_key(key: Any) -> str | None:
    if not isinstance(key, str):
        return None
    if not key.endswith("_csm"):
        return None
    domain = key[:-4].strip().lower()
    return domain if domain else None


def _coerce_csm_value(key: Any, value: Any, *, source: str) -> int | None:
    if isinstance(value, bool):
        LOG.warning("%s payload csm field %r has invalid bool value.", source, key)
        return None
    if isinstance(value, int):
        return value
    if isinstance(value, float) and value.is_integer():
        return int(value)
    if isinstance(value, str):
        text = value.strip()
        if text.isdigit():
            return int(text)
    LOG.warning("%s payload csm field %r has non-int value %r.", source, key, value)
    return None


def make_control_get_version_info_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("control","get_version_info") where payload is msg["control"]["get_version_info"].
    """

    def handler_control_get_version_info(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        control_obj = _as_mapping(msg.get("control"))
        if control_obj is None:
            return False

        payload = _as_mapping(control_obj.get("get_version_info"))
        if payload is None:
            return False

        error_code = payload.get("error_code")
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="control",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="control",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        outcome = _reconcile_control_get_version_info(state, payload, now=now())

        if outcome.changed_fields:
            emit(
                PanelVersionInfoUpdated(
                    kind=PanelVersionInfoUpdated.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    changed_fields=outcome.changed_fields,
                ),
                ctx,
            )

        if outcome.warnings:
            emit(
                DispatchRoutingError(
                    kind=DispatchRoutingError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    code="schema_warnings",
                    message="control.get_version_info payload contained type/schema warnings.",
                    keys=outcome.warnings,
                    severity="info",
                ),
                ctx,
            )

        return True

    return handler_control_get_version_info



"""Outbound request generators and registry."""

from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:

    class CommandSpec(Protocol):
        pass

    COMMANDS: dict[str, CommandSpec]


def __getattr__(name: str) -> Any:
    if name in {"COMMANDS", "CommandSpec"}:
        module = importlib.import_module(".registry", __name__)
        return getattr(module, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = ["CommandSpec", "COMMANDS"]



from __future__ import annotations

from elke27_lib.dispatcher import DispatchContext, PagedBlock
from elke27_lib.events import (
    ApiError,
    AuthorizationRequiredEvent,
    BootstrapCountsReady,
    CsmSnapshotUpdated,
    TableCsmChanged,
    TstatConfiguredInventoryReady,
    TstatConfiguredUpdated,
    TstatStatusUpdated,
    TstatTableInfoUpdated,
)
from elke27_lib.handlers import tstat as tstat_handler
from elke27_lib.states import PanelState, TstatState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, _ctx: DispatchContext) -> None:
        self.events.append(evt)


def _any_event(spy: _EmitSpy, kind: type) -> bool:
    return any(isinstance(evt, kind) for evt in spy.events)


def test_tstat_get_status_and_set_status_handlers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    get_status = tstat_handler.make_tstat_get_status_handler(state, emit, now=lambda: 1.0)
    set_status = tstat_handler.make_tstat_set_status_handler(state, emit, now=lambda: 2.0)

    assert get_status({"nope": {}}, make_ctx()) is False
    assert get_status({"tstat": {}}, make_ctx()) is False

    msg = {"tstat": {"get_status": {"error_code": 3}}}
    assert get_status(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {
        "tstat": {
            "get_status": {
                "tstat_id": 1,
                "temperature": 70,
                "mode": "HEAT",
                "fan_mode": "AUTO",
                "battery level": 95,
            }
        }
    }
    assert get_status(msg, make_ctx()) is True
    tstat = state.tstats[1]
    assert tstat.temperature == 70
    assert tstat.mode == "HEAT"
    assert tstat.fan_mode == "AUTO"
    assert tstat.battery_level == 95
    assert _any_event(emit, TstatStatusUpdated)

    emit.events.clear()
    msg = {"tstat": {"set_status": {"tstat_id": 1, "mode": "COOL", "cool_setpoint": 72}}}
    assert set_status(msg, make_ctx()) is True
    assert state.tstats[1].mode == "COOL"
    assert state.tstats[1].cool_setpoint == 72
    assert _any_event(emit, TstatStatusUpdated)


def test_tstat_configured_attribs_table_handlers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    configured = tstat_handler.make_tstat_get_configured_handler(state, emit, now=lambda: 3.0)
    attribs = tstat_handler.make_tstat_get_attribs_handler(state, emit, now=lambda: 4.0)
    table = tstat_handler.make_tstat_get_table_info_handler(state, emit, now=lambda: 5.0)

    msg = {"tstat": {"get_configured": {"error_code": 11008}}}
    assert configured(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"tstat": {"get_configured": {"tstats": [1, 2], "block_id": 1, "block_count": 1}}}
    assert configured(msg, make_ctx()) is True
    assert state.inventory.configured_tstats == {1, 2}
    assert state.inventory.configured_tstats_complete is True
    assert _any_event(emit, TstatConfiguredUpdated)
    assert _any_event(emit, TstatConfiguredInventoryReady)

    merge = tstat_handler.make_tstat_configured_merge(state)
    merged = merge(
        [
            PagedBlock(block_id=1, payload={"tstats": [1]}),
            PagedBlock(block_id=2, payload={"tstat_ids": [2]}),
        ],
        2,
    )
    assert merged == {"tstats": [1, 2], "block_count": 2}

    emit.events.clear()
    msg = {"tstat": {"get_attribs": {"tstat_id": 1, "name": " Upstairs ", "x": 1}}}
    assert attribs(msg, make_ctx()) is True
    assert state.tstats[1].name == "Upstairs"
    assert state.tstats[1].fields["x"] == 1

    emit.events.clear()
    state.table_info_known.update({"area", "zone", "output"})
    msg = {"tstat": {"get_table_info": {"table_elements": 1, "table_csm": "6"}}}
    assert table(msg, make_ctx()) is True
    assert _any_event(emit, TstatTableInfoUpdated)
    assert _any_event(emit, TableCsmChanged)
    assert _any_event(emit, CsmSnapshotUpdated)
    assert _any_event(emit, BootstrapCountsReady)


def test_tstat_helpers() -> None:
    tstat = TstatState(tstat_id=1)
    changed: set[str] = set()
    tstat_handler._apply_tstat_status_fields(
        tstat,
        {
            "temperature": 71,
            "cool_setpoint": 73,
            "heat_setpoint": 67,
            "mode": "HEAT",
            "fan_mode": "AUTO",
            "humidity": 40,
            "rssi": -55,
            "battery_level": 93,
            "prec": [1, 2],
            "extra": 1,
        },
        changed,
    )
    assert tstat.temperature == 71
    assert tstat.cool_setpoint == 73
    assert tstat.heat_setpoint == 67
    assert tstat.battery_level == 93
    assert tstat.prec == [1, 2]
    assert tstat.fields["extra"] == 1

    changed.clear()
    tstat_handler._apply_tstat_attribs(tstat, {"name": " Main ", "a": 1}, changed)
    assert tstat.name == "Main"
    assert tstat.fields["a"] == 1

    assert tstat_handler._extract_configured_tstat_ids({"tstats": [1, 2, 2]}) == [1, 2]
    assert tstat_handler._extract_configured_tstat_ids({"none": []}) == []
    assert tstat_handler._extract_table_csm({"table_csm": "3"}, domain="tstat") == 3



from __future__ import annotations

from elke27_lib.dispatcher import DispatchContext, PagedBlock
from elke27_lib.events import (
    ApiError,
    AuthorizationRequiredEvent,
    BootstrapCountsReady,
    CsmSnapshotUpdated,
    LightConfiguredInventoryReady,
    LightConfiguredUpdated,
    LightStatusUpdated,
    LightTableInfoUpdated,
    TableCsmChanged,
)
from elke27_lib.handlers import light as light_handler
from elke27_lib.states import LightState, PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, _ctx: DispatchContext) -> None:
        self.events.append(evt)


def _any_event(spy: _EmitSpy, kind: type) -> bool:
    return any(isinstance(evt, kind) for evt in spy.events)


def test_light_get_status_and_set_status_handlers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    get_status = light_handler.make_light_get_status_handler(state, emit, now=lambda: 1.0)
    set_status = light_handler.make_light_set_status_handler(state, emit, now=lambda: 2.0)

    assert get_status({"nope": {}}, make_ctx()) is False
    assert get_status({"light": {}}, make_ctx()) is False

    msg = {"light": {"get_status": {"error_code": 3}}}
    assert get_status(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"light": {"get_status": {"light_id": 1, "status": "on", "level": 45}}}
    assert get_status(msg, make_ctx()) is True
    light = state.lights[1]
    assert light.status == "ON"
    assert light.on is True
    assert light.level == 45
    assert _any_event(emit, LightStatusUpdated)

    emit.events.clear()
    msg = {"light": {"set_status": {"light_id": 1, "status": "OFF", "level": 0}}}
    assert set_status(msg, make_ctx()) is True
    assert state.lights[1].on is False
    assert _any_event(emit, LightStatusUpdated)


def test_light_configured_attribs_table_handlers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    configured = light_handler.make_light_get_configured_handler(state, emit, now=lambda: 3.0)
    attribs = light_handler.make_light_get_attribs_handler(state, emit, now=lambda: 4.0)
    table = light_handler.make_light_get_table_info_handler(state, emit, now=lambda: 5.0)

    msg = {"light": {"get_configured": {"error_code": 11008}}}
    assert configured(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"light": {"get_configured": {"lights": [1, 2], "block_id": 1, "block_count": 1}}}
    assert configured(msg, make_ctx()) is True
    assert state.inventory.configured_lights == {1, 2}
    assert _any_event(emit, LightConfiguredUpdated)
    assert _any_event(emit, LightConfiguredInventoryReady)

    merge = light_handler.make_light_configured_merge(state)
    merged = merge(
        [
            PagedBlock(block_id=1, payload={"lights": [1]}),
            PagedBlock(block_id=2, payload={"light_ids": [2]}),
        ],
        2,
    )
    assert merged == {"lights": [1, 2], "block_count": 2}

    emit.events.clear()
    msg = {"light": {"get_attribs": {"light_id": 1, "name": " Porch ", "area_id": 2}}}
    assert attribs(msg, make_ctx()) is True
    assert state.lights[1].name == "Porch"
    assert state.lights[1].area_id == 2

    emit.events.clear()
    state.table_info_known.update({"area", "zone", "output", "tstat"})
    msg = {"light": {"get_table_info": {"table_elements": 2, "table_csm": "7"}}}
    assert table(msg, make_ctx()) is True
    assert _any_event(emit, LightTableInfoUpdated)
    assert _any_event(emit, TableCsmChanged)
    assert _any_event(emit, CsmSnapshotUpdated)
    assert _any_event(emit, BootstrapCountsReady)


def test_light_helpers() -> None:
    light = LightState(light_id=1)
    light_handler._apply_light_status_fields(light, {"status": "off", "level": 0, "x": 1})
    assert light.status == "OFF"
    assert light.on is False
    assert light.level == 0
    assert light.fields["x"] == 1

    light_handler._apply_light_attribs(light, {"name": " A ", "area_id": 1, "y": 2})
    assert light.name == "A"
    assert light.area_id == 1
    assert light.fields["y"] == 2

    assert light_handler._extract_configured_ids({"lights": [1, 2, 2, 0]}, ("lights",)) == [1, 2]
    assert light_handler._extract_int({"table_elements": 1}, "table_elements") == 1
    assert light_handler._extract_int({"table_elements": "x"}, "table_elements") is None
    assert light_handler._extract_table_csm({"table_csm": "4"}, domain="light") == 4



from __future__ import annotations

from elke27_lib.const import E27ErrorCode
from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import ApiError, AreaStatusUpdated, Event
from elke27_lib.handlers.area import make_area_get_status_handler
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[Event] = []

    def __call__(self, evt: Event, _ctx: DispatchContext) -> None:
        self.events.append(evt)


_Ctx = make_ctx


def test_area_get_status_updates_state_and_emits_event() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_area_get_status_handler(state, emit, now=lambda: 123.0)

    msg = {
        "area": {
            "get_status": {
                "area_id": 1,
                "arm_state": "armed",
                "ready_status": "ready",
                "alarm_state": "none",
                "error_code": E27ErrorCode.ELKERR_NONE,
            }
        }
    }

    assert handler(msg, _Ctx()) is True
    area = state.areas[1]
    assert area.arm_state == "armed"
    assert area.ready_status == "ready"
    assert area.alarm_state == "none"
    assert any(isinstance(e, AreaStatusUpdated) for e in emit.events)


def test_area_get_status_error_code_emits_api_error() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_area_get_status_handler(state, emit, now=lambda: 123.0)

    msg = {"area": {"get_status": {"area_id": 1, "error_code": E27ErrorCode.ELKERR_INVALID_PARAM}}}

    assert handler(msg, _Ctx()) is True
    assert any(isinstance(e, ApiError) for e in emit.events)



# test/helpers/reporter.py
from __future__ import annotations

import importlib
import json
import os
import pathlib
import time
from dataclasses import dataclass, field
from typing import Any, cast


def _safe_json_default(obj: Any) -> Any:
    # Ensure JSON serialization never crashes reporting.
    if isinstance(obj, (bytes, bytearray)):
        return {"__bytes__": True, "len": len(obj), "sha256": None}
    try:
        return str(obj)
    except Exception:
        return repr(obj)


def _now_iso() -> str:
    # ISO-ish timestamp without timezone dependency.
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())


def _load_redactor():
    # Prefer project redaction helper if present; otherwise do minimal redaction.
    try:
        module = importlib.import_module("test.helpers.redaction")
        redact_record = getattr(module, "redact_record", None)
        if callable(redact_record):
            return redact_record
    except Exception:
        return None
    return None


_REDACT = _load_redactor()


def _minimal_redact(record: dict[str, Any]) -> dict[str, Any]:
    # Last-resort redaction: remove common secret-ish keys.
    secret_keys = {
        "pin",
        "pass",
        "access_code",
        "pass_phrase",
        "link_enc",
        "link_hmac",
        "sk",
        "shm",
        "session_key",
        "session_hmac",
        "key",
        "secret",
        "password",
    }

    def scrub(v: Any) -> Any:
        if isinstance(v, dict):
            v_map = cast(dict[str, Any], v)
            out: dict[str, Any] = {}
            for k, vv in v_map.items():
                if str(k).lower() in secret_keys:
                    out[k] = "<redacted>"
                else:
                    out[k] = scrub(vv)
            return out
        if isinstance(v, list):
            v_list = cast(list[Any], v)
            return [scrub(x) for x in v_list]
        return v

    return scrub(record)


@dataclass
class Reporter:
    run_id: str
    test_id: str
    artifacts_dir: pathlib.Path
    emit_jsonl: bool = True
    emit_yaml: bool = False
    enable: bool = True

    _t0: float = field(default_factory=time.monotonic, init=False)
    _records: list[dict[str, Any]] = field(default_factory=list, init=False)
    _jsonl_path: pathlib.Path | None = field(default=None, init=False)
    _yaml_path: pathlib.Path | None = field(default=None, init=False)

    def _t_ms(self) -> int:
        return int((time.monotonic() - self._t0) * 1000)

    def _ensure_paths(self) -> None:
        if not self.enable:
            return
        self.artifacts_dir.mkdir(parents=True, exist_ok=True)
        base = self.run_id.replace(os.sep, "_")
        self._jsonl_path = self.artifacts_dir / f"{base}.jsonl"
        self._yaml_path = self.artifacts_dir / f"{base}.summary.yaml"

    def _apply_redaction(self, record: dict[str, Any]) -> dict[str, Any]:
        if _REDACT is not None:
            try:
                redacted = _REDACT(record)
                if isinstance(redacted, dict):
                    return cast(dict[str, Any], redacted)
            except Exception:
                # If project redactor fails, fall back.
                return _minimal_redact(record)
        return _minimal_redact(record)

    def emit(self, record_type: str, **fields: Any) -> dict[str, Any]:
        """
        Emit a typed record. Always adds:
          - run_id
          - test_id
          - record_type
          - t_ms
          - ts_utc
        Writes JSONL immediately (append-only) when enabled.
        """
        if not self.enable:
            return {}

        self._ensure_paths()

        record: dict[str, Any] = {
            "run_id": self.run_id,
            "test_id": self.test_id,
            "record_type": record_type,
            "t_ms": self._t_ms(),
            "ts_utc": _now_iso(),
        }
        record.update(fields)

        record = self._apply_redaction(record)

        self._records.append(record)

        if self.emit_jsonl and self._jsonl_path is not None:
            with self._jsonl_path.open("a", encoding="utf-8") as f:
                f.write(json.dumps(record, default=_safe_json_default, sort_keys=False))
                f.write("\n")

        return record

    def test_start(self) -> None:
        self.emit(
            "test_start",
            meta={
                "pid": os.getpid(),
            },
        )

    def test_end(self, outcome: str, when: str | None, longrepr: Any) -> None:
        err: dict[str, Any] | None = None
        if longrepr is not None:
            # pytest longrepr can be various types; store as string safely.
            err = {"when": when, "longrepr": str(longrepr)}

        self.emit(
            "test_end",
            outcome=outcome,
            error=err,
        )

    def finalize(self) -> None:
        """
        Finalize reporting. JSONL is already written incrementally.
        YAML summary is derived from the collected in-memory records.
        """
        if not self.enable:
            return

        self._ensure_paths()

        if self.emit_yaml and self._yaml_path is not None:
            summary = {
                "run_id": self.run_id,
                "test_id": self.test_id,
                "record_count": len(self._records),
                "records": self._records,
            }
            # YAML is optional; try PyYAML, otherwise write a JSON-formatted fallback.
            try:
                yaml_mod = importlib.import_module("yaml")
                safe_dump = getattr(yaml_mod, "safe_dump", None)
                if callable(safe_dump):
                    with self._yaml_path.open("w", encoding="utf-8") as f:
                        safe_dump(summary, f, sort_keys=False, default_flow_style=False)
                    return
            except Exception:
                pass
            with self._yaml_path.open("w", encoding="utf-8") as f:
                # Fallback: write JSON with .yaml extension (still readable, avoids hard dependency)
                f.write(json.dumps(summary, indent=2, default=_safe_json_default, sort_keys=False))
                f.write("\n")



"""Message builders used in tests and offline tooling."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from elke27_lib.features.area import build_area_get_status_payload
from elke27_lib.features.control import build_control_authenticate_payload


def _wrap_request(
    *,
    seq: int,
    domain: str,
    name: str,
    payload: Mapping[str, Any],
    session_id: int | None = None,
) -> dict[str, Any]:
    msg: dict[str, Any] = {"seq": seq}
    if session_id is not None:
        msg["session_id"] = session_id
    msg[domain] = {name: dict(payload)}
    return msg


def build_area_get_status(*, seq: int, session_id: int, area_id: int) -> dict[str, Any]:
    payload = build_area_get_status_payload(area_id=area_id)
    return _wrap_request(
        seq=seq, session_id=session_id, domain="area", name="get_status", payload=payload
    )


def build_authenticate(*, seq: int, pin: int | str) -> dict[str, Any]:
    payload = build_control_authenticate_payload(pin=pin)
    return _wrap_request(
        seq=seq, session_id=None, domain="control", name="authenticate", payload=payload
    )


__all__ = ["build_area_get_status", "build_authenticate"]



from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Any, cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.errors import E27Timeout
from test.helpers.internal import get_kernel, get_private


class _FakeSession:
    cfg: object

    def __init__(self) -> None:
        self.cfg = type("_Cfg", (), {"host": "test-host", "port": 1})()
        self.sent: list[dict[str, Any]] = []

    def send_json(
        self,
        msg: dict[str, Any],
        *,
        priority: object = None,
        on_sent: Callable[[float], None] | None = None,
        on_fail: Callable[[BaseException], None] | None = None,
    ) -> None:
        del priority, on_fail
        self.sent.append(msg)
        if on_sent is not None:
            on_sent(0.0)


def _set_session(kernel: object, session: _FakeSession) -> None:
    cast(Any, kernel)._session = session


@pytest.mark.asyncio
async def test_async_execute_sends_seq_and_resolves() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    fake_session = _FakeSession()
    _set_session(kernel, fake_session)
    kernel.state.panel.session_id = 1

    task = asyncio.create_task(client.async_execute("control_get_version_info"))
    await asyncio.sleep(0)

    assert fake_session.sent
    sent = fake_session.sent[0]
    seq = sent["seq"]
    assert seq > 0

    on_message = get_private(kernel, "_on_message")
    on_message({"seq": seq, "control": {"get_version_info": {"version": "1.0"}}})
    result = await task

    assert result.ok is True
    assert result.data == {"version": "1.0"}
    pending = get_private(kernel, "_pending_responses")
    assert pending.pending_count() == 0


@pytest.mark.asyncio
async def test_async_execute_ignores_broadcast() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    fake_session = _FakeSession()
    _set_session(kernel, fake_session)
    kernel.state.panel.session_id = 1

    task = asyncio.create_task(client.async_execute("control_get_version_info"))
    await asyncio.sleep(0)

    sent = fake_session.sent[0]
    seq = sent["seq"]

    on_message = get_private(kernel, "_on_message")
    on_message({"seq": 0, "control": {"get_version_info": {"version": "ignored"}}})
    await asyncio.sleep(0)
    assert not task.done()

    on_message = get_private(kernel, "_on_message")
    on_message({"seq": seq, "control": {"get_version_info": {"version": "1.1"}}})
    result = await task

    assert result.ok is True
    assert result.data == {"version": "1.1"}
    pending = get_private(kernel, "_pending_responses")
    assert pending.pending_count() == 0


@pytest.mark.asyncio
async def test_async_execute_times_out_and_cleans_pending() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    fake_session = _FakeSession()
    _set_session(kernel, fake_session)
    kernel.state.panel.session_id = 1

    result = await client.async_execute("control_get_version_info", timeout_s=0.01)

    assert result.ok is False
    assert isinstance(result.error, E27Timeout)
    error_text = str(result.error)
    assert "control_get_version_info" in error_text or "seq=" in error_text
    pending = get_private(kernel, "_pending_responses")
    assert pending.pending_count() == 0



#!/usr/bin/env python3
"""
e27_live_version_info.py

Live example: discover -> link -> connect (HELLO) -> control.get_version_info -> print PanelState.

NOTE:
- This example performs provisioning (API_LINK) via elk.link() so it can be run standalone.
- If you want “no provisioning” runs, we should add a dedicated Elk method to set context
  from an existing link key without calling link(). (Not implemented yet.)

Environment variables (optional):
  ELKE27_HOST, ELKE27_PORT
  ELKE27_MN, ELKE27_SN, ELKE27_FWVER, ELKE27_HWVER, ELKE27_OSVER
  ELKE27_ACCESS_CODE, ELKE27_PASSPHRASE
  ELK_HOST, ELK_PORT
  ELK_MN, ELK_SN, ELK_FWVER, ELK_HWVER, ELK_OSVER
  ELK_ACCESS_CODE, ELK_PASSPHRASE
"""

from __future__ import annotations

import argparse
import asyncio
import logging
from dataclasses import dataclass

from elke27_lib import Elk, linking
from elke27_lib.session import SessionState

LOG = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class E27Credentials:
    access_code: str
    passphrase: str


def _env(name: str, default: str | None = None) -> str | None:
    import os
    return os.environ.get(name, default)


def _env_any(names: list[str], default: str | None = None) -> str | None:
    for name in names:
        value = _env(name)
        if value:
            return value
    return default


def _truthy(value: str | None) -> bool:
    if value is None:
        return False
    return str(value).strip().lower() in {"1", "true", "yes", "on"}


async def main_async() -> int:
    ap = argparse.ArgumentParser()
    ap.add_argument("--timeout", type=float, default=10.0, help="Overall timeout for request/response")
    ap.add_argument("--pump-s", type=float, default=0.25, help="Per-pump timeout")
    ap.add_argument("--address", type=str, default=None, help="Optional discovery broadcast address override")
    ap.add_argument("--log-level", default=_env_any(["LOG_LEVEL"], "INFO"), help="DEBUG/INFO/WARNING/ERROR")
    ap.add_argument(
        "--wire-log",
        action="store_true",
        default=_truthy(_env_any(["ELKE27_WIRE_LOG", "ELK_WIRE_LOG"], None)),
        help="Enable raw RX/TX hex dump logging",
    )
    ap.add_argument("--pin", type=int, default=None, help="Optional pin for authenticate diagnostic")
    ap.add_argument("--auth-only", action="store_true", help="Exit after authenticate diagnostic reply")
    ap.add_argument(
        "--host",
        type=str,
        default=_env_any(["ELKE27_HOST", "ELK_HOST"], None),
        help="Panel IP/host (skip discovery)",
    )
    ap.add_argument(
        "--port",
        type=int,
        default=int(_env_any(["ELKE27_PORT", "ELK_PORT"], "2101") or 2101),
        help="Panel port (default 2101)",
    )

    ap.add_argument("--mn", default=_env_any(["ELKE27_MN", "ELK_MN"], ""), help="Manufacturer number (MN)")
    ap.add_argument("--sn", default=_env_any(["ELKE27_SN", "ELK_SN"], ""), help="Serial number (SN)")
    ap.add_argument(
        "--fwver",
        default=_env_any(["ELKE27_FWVER", "ELK_FWVER"], "0"),
        help="Firmware version",
    )
    ap.add_argument(
        "--hwver",
        default=_env_any(["ELKE27_HWVER", "ELK_HWVER"], "0"),
        help="Hardware version",
    )
    ap.add_argument(
        "--osver",
        default=_env_any(["ELKE27_OSVER", "ELK_OSVER"], "0"),
        help="OS version",
    )

    ap.add_argument(
        "--access-code",
        default=_env_any(["ELKE27_ACCESS_CODE", "ELK_ACCESS_CODE"], ""),
        help="Access code",
    )
    ap.add_argument(
        "--passphrase",
        default=_env_any(["ELKE27_PASSPHRASE", "ELK_PASSPHRASE"], ""),
        help="Provisioning passphrase",
    )
    args = ap.parse_args()
    logging.basicConfig(level=getattr(logging, str(args.log_level).upper(), logging.INFO))

    if not args.mn or not args.sn:
        raise SystemExit("ERROR: client_identity requires --mn and --sn (or ELK_MN / ELK_SN).")
    if not args.access_code or not args.passphrase:
        raise SystemExit("ERROR: credentials require --access-code and --passphrase (or env vars).")

    # Panel selection
    if args.host:
        panel = {"host": args.host, "port": args.port}
    else:
        discovered = await Elk.discover(timeout=int(args.timeout), address=args.address)
        if not discovered.panels:
            raise SystemExit("ERROR: no panels discovered. Provide --host/--port or fix discovery environment.")
        panel = discovered.panels[0]

    elk = Elk()

    client_identity = linking.E27Identity(
        mn=str(args.mn),
        sn=str(args.sn),
        fwver=str(args.fwver),
        hwver=str(args.hwver),
        osver=str(args.osver),
    )
    creds = E27Credentials(access_code=str(args.access_code), passphrase=str(args.passphrase))

    link_keys = await elk.link(panel, client_identity, creds, timeout_s=float(args.timeout))

    if isinstance(panel, dict):
        panel_host = panel.get("panel_host") or panel.get("host") or panel.get("ip_address")
        panel_port = int(panel.get("port") or 2101)
    else:
        panel_host = panel.panel_host
        panel_port = int(panel.port)

    from elke27_lib.session import SessionConfig

    session_cfg = SessionConfig(
        host=panel_host,
        port=panel_port,
        hello_timeout_s=float(args.timeout),
        wire_log=bool(args.wire_log),
    )
    state = await elk.connect(link_keys, client_identity=client_identity, session_config=session_cfg)
    print(f"Session state: {state}")
    if state is not SessionState.ACTIVE:
        raise SystemExit(f"ERROR: session not ACTIVE (state={state})")

    if args.pin is not None:
        auth_msg = {"seq": 110, "authenticate": {"pin": int(args.pin)}}
        elk.session.send_json(auth_msg)
        auth_reply = elk.session.recv_json(timeout_s=2.0)
        print(f"Authenticate reply: {auth_reply}")
        if args.auth_only:
            await elk.close()
            return 0

    # seq = elk.request(("control", "get_version_info"))
    # print(f"Sent control.get_version_info seq={seq}")

    # deadline = time.monotonic() + float(args.timeout)

    # while time.monotonic() < deadline:
    #     # Drive inbound receive -> dispatch -> events
    #     elk.session.pump_once(timeout_s=float(args.pump_s))

    #     for evt in elk.drain_events():
    #         if evt.kind == PanelVersionInfoUpdated.KIND:
    #             print("\nPanel version info updated:")
    #             print(f"  session_id     = {elk.state.panel.session_id}")
    #             print(f"  connected      = {elk.state.panel.connected}")
    #             print(f"  last_message_at= {elk.state.panel.last_message_at}")
    #             print(f"  model          = {elk.state.panel.model}")
    #             print(f"  firmware       = {elk.state.panel.firmware}")
    #             print(f"  serial         = {elk.state.panel.serial}")
    #             await elk.close()
    #             return 0

    #     await asyncio.sleep(0)  # yield

    # print("Timed out waiting for PanelVersionInfoUpdated event.")
    await elk.close()
    return 0


def main() -> int:
    return asyncio.run(main_async())


if __name__ == "__main__":
    raise SystemExit(main())



"""Control domain request generators."""

from __future__ import annotations

ResponseKey = tuple[str, str]


def generator_control_get_version_info() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("control", "get_version_info")


def generator_control_get_table_info() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("control", "get_table_info")


def generator_control_get_trouble() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("control", "get_trouble")


def generator_control_authenticate(*, pin: int) -> tuple[dict[str, object], ResponseKey]:
    if not (0 <= pin <= 999999):
        raise ValueError("pin must be in range 0..999999")
    return {"pin": pin}, ("control", "authenticate")



from __future__ import annotations

import asyncio
import types
from types import SimpleNamespace
from typing import Any, cast

import pytest

import elke27_lib.kernel as kernel_mod
from elke27_lib.errors import ConnectionLost, E27Error, E27Timeout
from elke27_lib.events import ConnectionStateChanged
from elke27_lib.kernel import E27Kernel, KernelError, KernelMissingContextError
from elke27_lib.linking import E27Identity, E27LinkKeys
from elke27_lib.outbound import OutboundPriority
from elke27_lib.session import SessionIOError, SessionProtocolError, SessionState


def test_as_mapping_and_redact_value(monkeypatch: pytest.MonkeyPatch) -> None:
    assert kernel_mod._as_mapping({"a": 1}) is not None
    assert kernel_mod._as_mapping(1) is None

    monkeypatch.setattr(kernel_mod, "REDACT_DIAGNOSTICS", True)
    redacted = kernel_mod._redact_value([{"pin": "123"}])
    assert redacted == [{"pin": "***"}]
    redacted_tuple = kernel_mod._redact_value(({"pin": "123"},))
    assert redacted_tuple == ({"pin": "***"},)


def test_session_property_raises() -> None:
    kernel = E27Kernel()
    with pytest.raises(KernelError):
        _ = kernel.session


@pytest.mark.asyncio
async def test_discover_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    class _Scanner:
        async def async_scan(self, *_a: Any, **_k: Any) -> list[object]:
            raise RuntimeError("boom")

    monkeypatch.setattr(kernel_mod.discovery, "AIOELKDiscovery", _Scanner)
    with pytest.raises(KernelError):
        await E27Kernel.discover()

    class _Scanner2:
        async def async_scan(self, *_a: Any, **_k: Any) -> list[object]:
            return [object()]

    monkeypatch.setattr(kernel_mod.discovery, "AIOELKDiscovery", _Scanner2)
    with pytest.raises(KernelError):
        await E27Kernel.discover()


@pytest.mark.asyncio
async def test_connect_missing_context() -> None:
    kernel = E27Kernel()
    link_keys = E27LinkKeys(tempkey_hex="aa", linkkey_hex="bb", linkhmac_hex="cc")
    identity = E27Identity(mn="m", sn="s", fwver="f", hwver="h", osver="o")
    with pytest.raises(KernelMissingContextError):
        await kernel.connect(link_keys, client_identity=identity, session_config=None, panel=None)

    bad_identity = E27Identity(mn="", sn="s", fwver="f", hwver="h", osver="o")
    with pytest.raises(KernelMissingContextError):
        await kernel.connect(
            link_keys,
            client_identity=bad_identity,
            session_config=kernel_mod.session_mod.SessionConfig(host="h", port=1),
        )


@pytest.mark.asyncio
async def test_close_paths(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    kernel = E27Kernel()
    await kernel.close()

    class _Session:
        info = SimpleNamespace(session_id=1)
        state = SessionState.ACTIVE

        def close(self) -> None:
            raise OSError("boom")

    cast(Any, kernel)._session = _Session()
    caplog.set_level("DEBUG", logger="elke27_lib.kernel")
    await kernel.close()


def test_start_stop_keepalive(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    cast(Any, kernel)._loop = None
    kernel._start_keepalive()

    class _Task:
        def done(self) -> bool:
            return False

        def cancel(self) -> None:
            return None

    def _create_task(coro: Any) -> _Task:
        coro.close()
        return _Task()

    cast(Any, kernel)._loop = SimpleNamespace(create_task=_create_task)
    cast(Any, kernel)._keepalive_task = None
    kernel._start_keepalive()
    cast(Any, kernel)._keepalive_task = _Task()
    kernel._start_keepalive()

    cast(Any, kernel)._keepalive_task = None
    kernel._stop_keepalive()
    cast(Any, kernel)._keepalive_task = _Task()
    kernel._stop_keepalive()


@pytest.mark.asyncio
async def test_keepalive_loop_branches(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    kernel._keepalive_enabled = True
    kernel._closing = True
    await kernel._keepalive_loop()

    kernel._closing = False
    kernel._keepalive_enabled = False
    await kernel._keepalive_loop()

    async def _sleep_cancel(_delay: float) -> None:
        raise asyncio.CancelledError()

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_cancel)
    kernel._keepalive_enabled = True
    kernel._last_exchange_at = kernel.now()
    await kernel._keepalive_loop()

    kernel._keepalive_interval_s = 0.0
    kernel._last_exchange_at = 0.0
    cast(Any, kernel)._session = None
    await kernel._keepalive_loop()

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    await kernel._keepalive_loop()

    kernel._request_state = kernel_mod._RequestState.IDLE
    cast(Any, kernel)._session = SimpleNamespace(
        state=SessionState.ACTIVE, _outbound=SimpleNamespace(is_idle=lambda: False)
    )
    await kernel._keepalive_loop()

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._keepalive_inflight = True
    await kernel._keepalive_loop()

    kernel._keepalive_inflight = False

    async def _send_keepalive() -> bool:
        kernel._closing = True
        return True

    monkeypatch.setattr(kernel, "_send_keepalive_request", _send_keepalive)
    kernel._closing = False
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_keepalive_loop_sleep_cancellations(monkeypatch: pytest.MonkeyPatch) -> None:
    async def _sleep_cancel(_delay: float) -> None:
        raise asyncio.CancelledError()

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_cancel)

    kernel = E27Kernel()
    kernel._keepalive_enabled = True
    kernel._last_exchange_at = kernel.now()
    kernel._keepalive_interval_s = 100.0
    await kernel._keepalive_loop()

    kernel._keepalive_interval_s = 0.0
    kernel._last_exchange_at = 0.0
    cast(Any, kernel)._session = None
    await kernel._keepalive_loop()

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    await kernel._keepalive_loop()

    kernel._request_state = kernel_mod._RequestState.IDLE
    cast(Any, kernel)._session = SimpleNamespace(
        state=SessionState.ACTIVE, _outbound=SimpleNamespace(is_idle=lambda: False)
    )
    await kernel._keepalive_loop()

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._keepalive_inflight = True
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_keepalive_loop_wait_cancel_line_659(monkeypatch: pytest.MonkeyPatch) -> None:
    async def _sleep_cancel(_delay: float) -> None:
        raise asyncio.CancelledError()

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_cancel)
    kernel = E27Kernel()
    kernel._keepalive_enabled = True
    kernel._last_exchange_at = kernel.now()
    kernel._keepalive_interval_s = 100.0
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_keepalive_loop_session_cancel_line_666(monkeypatch: pytest.MonkeyPatch) -> None:
    async def _sleep_cancel(_delay: float) -> None:
        raise asyncio.CancelledError()

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_cancel)
    kernel = E27Kernel()
    kernel._keepalive_enabled = True
    kernel._keepalive_interval_s = 0.0
    kernel._last_exchange_at = 0.0
    cast(Any, kernel)._session = None
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_keepalive_loop_request_cancel_line_676(monkeypatch: pytest.MonkeyPatch) -> None:
    async def _sleep_cancel(_delay: float) -> None:
        raise asyncio.CancelledError()

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_cancel)
    kernel = E27Kernel()
    kernel._keepalive_enabled = True
    kernel._keepalive_interval_s = 0.0
    kernel._last_exchange_at = 0.0
    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_keepalive_loop_outbound_cancel_line_683(monkeypatch: pytest.MonkeyPatch) -> None:
    async def _sleep_cancel(_delay: float) -> None:
        raise asyncio.CancelledError()

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_cancel)
    kernel = E27Kernel()
    kernel._keepalive_enabled = True
    kernel._keepalive_interval_s = 0.0
    kernel._last_exchange_at = 0.0
    cast(Any, kernel)._session = SimpleNamespace(
        state=SessionState.ACTIVE, _outbound=SimpleNamespace(is_idle=lambda: False)
    )
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_keepalive_loop_inflight_cancel_line_689(monkeypatch: pytest.MonkeyPatch) -> None:
    async def _sleep_cancel(_delay: float) -> None:
        raise asyncio.CancelledError()

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_cancel)
    kernel = E27Kernel()
    kernel._keepalive_enabled = True
    kernel._keepalive_interval_s = 0.0
    kernel._last_exchange_at = 0.0
    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._keepalive_inflight = True
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_keepalive_loop_continue_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    kernel._keepalive_enabled = True

    async def _sleep_noop(_delay: float) -> None:
        kernel._closing = True

    monkeypatch.setattr(kernel_mod.asyncio, "sleep", _sleep_noop)

    kernel._closing = False
    kernel._last_exchange_at = kernel.now()
    kernel._keepalive_interval_s = 100.0
    await kernel._keepalive_loop()

    kernel._closing = False
    kernel._keepalive_interval_s = 0.0
    kernel._last_exchange_at = 0.0
    cast(Any, kernel)._session = None
    await kernel._keepalive_loop()

    kernel._closing = False
    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    await kernel._keepalive_loop()

    kernel._closing = False
    kernel._request_state = kernel_mod._RequestState.IDLE
    cast(Any, kernel)._session = SimpleNamespace(
        state=SessionState.ACTIVE, _outbound=SimpleNamespace(is_idle=lambda: False)
    )
    await kernel._keepalive_loop()

    kernel._closing = False
    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._keepalive_inflight = True
    await kernel._keepalive_loop()


@pytest.mark.asyncio
async def test_send_keepalive_request_branches(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    monkeypatch.setattr(kernel, "_set_loop_if_needed", lambda: None)
    assert await kernel._send_keepalive_request() is False

    cast(Any, kernel)._loop = asyncio.get_running_loop()
    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    kernel._last_exchange_at = kernel.now()
    kernel._keepalive_interval_s = 10.0
    assert await kernel._send_keepalive_request() is True

    kernel._keepalive_inflight = True
    kernel._last_exchange_at = 0.0
    kernel._keepalive_interval_s = 0.0
    assert await kernel._send_keepalive_request() is True
    kernel._keepalive_inflight = False

    kernel.requests.register(("system", "r_u_alive"), lambda: {})

    def _send_raise(*_a: Any, **_k: Any) -> int:
        raise RuntimeError("boom")

    monkeypatch.setattr(kernel, "_send_request_with_seq", _send_raise)
    kernel._last_exchange_at = 0.0
    kernel._keepalive_interval_s = 0.0
    assert await kernel._send_keepalive_request() is False

    def _send_timeout(seq: int, *_a: Any, **_k: Any) -> int:
        kernel._signal_sent_event(seq)
        kernel._pending_responses.fail(seq, E27Timeout("t"))
        return seq

    monkeypatch.setattr(kernel, "_send_request_with_seq", _send_timeout)
    assert await kernel._send_keepalive_request() in {True, False}

    def _send_fail(seq: int, *_a: Any, **_k: Any) -> int:
        kernel._signal_sent_event(seq)
        kernel._pending_responses.fail(seq, RuntimeError("boom"))
        return seq

    monkeypatch.setattr(kernel, "_send_request_with_seq", _send_fail)
    assert await kernel._send_keepalive_request() in {True, False}


def test_load_features_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    called = {"load": 0}

    def _fake_blocking(mods: Any) -> None:
        called["load"] += 1

    monkeypatch.setattr(kernel, "load_features_blocking", _fake_blocking)
    kernel.load_features(["x"])
    assert called["load"] == 1

    fake_mod = types.ModuleType("fake")
    monkeypatch.setattr(kernel_mod.importlib, "import_module", lambda name: fake_mod)
    kernel2 = E27Kernel()
    with pytest.raises(RuntimeError):
        kernel2.load_features_blocking(["fake"])


def test_bootstrap_requests_and_csm_refresh(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    kernel._bootstrap_requests()

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    monkeypatch.setattr(kernel.requests, "get", lambda *_a, **_k: True)
    monkeypatch.setattr(kernel, "request", lambda *_a, **_k: (_ for _ in ()).throw(E27Error("x")))
    kernel._bootstrap_requests()

    cast(Any, kernel)._session = None
    with pytest.raises(KernelError):
        kernel.request_csm_refresh()

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE)
    monkeypatch.setattr(
        kernel.requests,
        "get",
        lambda route: True if route == ("control", "authenticate") else None,
    )
    monkeypatch.setattr(kernel, "request", lambda *_a, **_k: (_ for _ in ()).throw(E27Error("x")))
    kernel.request_csm_refresh(auth_pin=1234, domains=["area"])

    monkeypatch.setattr(kernel.requests, "get", lambda *_a, **_k: True)
    kernel.request_csm_refresh(auth_pin=1234, domains=["area"])


def test_on_message_paths(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    kernel = E27Kernel()
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    kernel._active_seq = 99
    caplog.set_level("DEBUG", logger="elke27_lib.kernel")
    kernel._on_message(types.MappingProxyType({"authenticate": {"seq": 1, "session_id": 2}}))
    kernel._on_message(
        types.MappingProxyType({"seq": "x", "authenticate": {"seq": 2, "session_id": 3}})
    )
    assert kernel.state.panel.session_id == 3

    kernel._on_message({"seq": 1, "session_id": 2})


def test_on_idle_and_disconnect(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    called = {"try": 0, "abort": 0, "paged": 0}
    monkeypatch.setattr(kernel, "_try_send_next", lambda: called.__setitem__("try", 1))
    kernel._on_idle()
    assert called["try"] == 1

    kernel._closing = True
    kernel._on_session_disconnected(RuntimeError("x"))

    kernel._closing = False
    kernel._closed_explicitly = True
    kernel._on_session_disconnected(SessionIOError("x"))

    kernel._closed_explicitly = False
    monkeypatch.setattr(kernel, "_abort_requests", lambda *_a, **_k: called.__setitem__("abort", 1))
    monkeypatch.setattr(
        kernel.dispatcher, "abort_paged_transfers", lambda *_a, **_k: called.__setitem__("paged", 1)
    )
    kernel._on_session_disconnected(SessionProtocolError("x"))
    assert called["abort"] == 1
    assert called["paged"] == 1


def test_is_valid_attrib_id() -> None:
    kernel = E27Kernel()
    assert kernel._is_valid_attrib_id("area", 0) is False
    inv = kernel.state.inventory
    inv.area_discovery_max_id = 1
    assert kernel._is_valid_attrib_id("area", 2) is False
    inv.configured_areas = {1}
    assert kernel._is_valid_attrib_id("area", 2) is False
    inv.zone_discovery_max_id = 1
    assert kernel._is_valid_attrib_id("zone", 2) is False
    inv.zone_discovery_max_id = None
    inv.configured_zones = {1}
    assert kernel._is_valid_attrib_id("zone", 2) is False
    inv.configured_outputs = {1}
    assert kernel._is_valid_attrib_id("output", 2) is False
    inv.configured_users = {1}
    assert kernel._is_valid_attrib_id("user", 2) is False
    inv.configured_keypads = {1}
    assert kernel._is_valid_attrib_id("keypad", 2) is False
    kernel.state.table_info_by_domain["area"] = {"table_elements": 1}
    inv.area_discovery_max_id = None
    inv.configured_areas = set()
    assert kernel._is_valid_attrib_id("area", 2) is False


def test_signal_sent_event_exception_path() -> None:
    kernel = E27Kernel()

    class _Event:
        def __init__(self) -> None:
            self._loop = None
            self.calls = 0

        def set(self) -> None:
            self.calls += 1
            if self.calls == 1:
                raise RuntimeError("boom")

    event = _Event()
    with kernel._sent_event_lock:
        kernel._sent_events[1] = event  # type: ignore[assignment]
    kernel._signal_sent_event(1)


def test_enqueue_kick_and_try_send_next(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    item = kernel_mod._QueuedRequest(
        seq=1,
        domain="test",
        name="cmd",
        payload={},
        pending=True,
        opaque=None,
        expected_route=None,
        priority=OutboundPriority.HIGH,
        timeout_s=1.0,
    )
    called = {"try": 0}
    monkeypatch.setattr(kernel, "_try_send_next", lambda: called.__setitem__("try", 1))
    kernel._enqueue_request(item)
    assert kernel._request_queue_high

    cast(Any, kernel)._loop = SimpleNamespace(is_running=lambda: False)
    kernel._kick_scheduler()
    assert called["try"] == 1

    class _Loop:
        def is_running(self) -> bool:
            return True

        def call_soon_threadsafe(self, fn: Any) -> None:
            fn()

    cast(Any, kernel)._loop = _Loop()
    kernel._kick_scheduler()

    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    kernel._try_send_next()
    kernel._request_state = kernel_mod._RequestState.IDLE
    kernel._request_queue_high.clear()
    kernel._request_queue_normal.clear()
    kernel._try_send_next()
    cast(Any, kernel)._session = None
    kernel._request_queue_high.append(item)
    kernel._try_send_next()
    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.DISCONNECTED)
    kernel._request_queue_high.append(item)
    kernel._try_send_next()


def test_try_send_next_session_not_active() -> None:
    kernel = E27Kernel()
    kernel._request_state = kernel_mod._RequestState.IDLE
    item = kernel_mod._QueuedRequest(
        seq=1,
        domain="test",
        name="cmd",
        payload={},
        pending=True,
        opaque=None,
        expected_route=None,
        priority=OutboundPriority.NORMAL,
        timeout_s=1.0,
    )
    kernel._request_queue_high.append(item)
    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.DISCONNECTED)
    kernel._try_send_next()


def test_try_send_next_paged_and_send_fail(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    cast(Any, kernel)._session = SimpleNamespace(
        state=SessionState.ACTIVE, send_json=lambda *_a, **_k: None
    )
    kernel.dispatcher.is_paged = lambda *_a, **_k: True  # type: ignore[assignment]
    kernel.dispatcher.add_pending = lambda *_a, **_k: None  # type: ignore[assignment]
    kernel.state.panel.session_id = 1

    item = kernel_mod._QueuedRequest(
        seq=1,
        domain="zone",
        name="get_configured",
        payload={"block_id": 1},
        pending=True,
        opaque=None,
        expected_route=("zone", "get_configured"),
        priority=OutboundPriority.NORMAL,
        timeout_s=1.0,
    )
    kernel._request_queue_high.append(item)

    def _send_json(*_a: Any, **_k: Any) -> None:
        raise RuntimeError("boom")

    kernel._session.send_json = _send_json  # type: ignore[assignment]
    kernel._log.setLevel("DEBUG")
    kernel._try_send_next()


def test_reply_timeout_and_abort(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    kernel = E27Kernel()
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    kernel._active_seq = 1
    kernel._active_request = None
    caplog.set_level("WARNING", logger="elke27_lib.kernel")
    kernel._on_reply_timeout(1)

    kernel._active_request = kernel_mod._QueuedRequest(
        seq=1,
        domain="zone",
        name="get_status",
        payload={},
        pending=True,
        opaque=None,
        expected_route=("zone", "get_status"),
        priority=OutboundPriority.NORMAL,
        timeout_s=1.0,
    )
    kernel.dispatcher.drop_pending = lambda *_a, **_k: None  # type: ignore[assignment]
    kernel._pending_responses.fail = lambda *_a, **_k: None  # type: ignore[assignment]
    kernel._on_reply_timeout(1)

    kernel._active_seq = 2
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    kernel._active_released = False
    kernel._handle_send_failure(1, RuntimeError("boom"))

    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    kernel._active_seq = 3
    kernel._active_released = False
    kernel._request_queue_high.append(
        kernel_mod._QueuedRequest(
            seq=4,
            domain="zone",
            name="get_status",
            payload={},
            pending=True,
            opaque=None,
            expected_route=("zone", "get_status"),
            priority=OutboundPriority.NORMAL,
            timeout_s=1.0,
        )
    )
    kernel._abort_requests(ConnectionLost("x"))

    kernel._active_released = True
    kernel._complete_active(reason="noop")


def test_mark_send_failed_request_and_next_seq(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    kernel._pending_responses.fail = lambda *_a, **_k: None  # type: ignore[assignment]
    kernel.dispatcher.drop_pending = lambda *_a, **_k: None  # type: ignore[assignment]
    kernel._mark_send_failed(1, RuntimeError("x"))

    kernel.requests.require = lambda *_a, **_k: (lambda **_kw: {"ok": True})  # type: ignore[assignment]
    monkeypatch.setattr(kernel, "_send_request", lambda *_a, **_k: 5)
    assert kernel.request(("zone", "get_status")) == 5

    kernel._seq = 0
    assert kernel._next_seq() == 10

    cast(Any, kernel)._active_timeout_handle = SimpleNamespace(cancel=lambda: None)
    cast(Any, kernel)._loop = SimpleNamespace(call_later=lambda *_a, **_k: None)
    kernel._arm_reply_timeout(1, 0)
    kernel._arm_reply_timeout(1, 1)
    kernel._request_state = kernel_mod._RequestState.IN_FLIGHT
    kernel._active_seq = 1
    kernel._active_released = True
    kernel._on_reply_timeout(1)


def test_send_request_with_seq_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    with pytest.raises(KernelError):
        kernel._send_request_with_seq(
            1, "zone", "get_status", {}, pending=True, opaque=None, expected_route=None
        )

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.DISCONNECTED)
    with pytest.raises(KernelError):
        kernel._send_request_with_seq(
            1, "zone", "get_status", {}, pending=True, opaque=None, expected_route=None
        )

    def _send_json(*_a: Any, **_k: Any) -> None:
        raise RuntimeError("boom")

    cast(Any, kernel)._session = SimpleNamespace(state=SessionState.ACTIVE, send_json=_send_json)
    with pytest.raises(KernelError):
        kernel._send_request_with_seq(
            1,
            "zone",
            "get_status",
            {"x": 1},
            pending=False,
            opaque=None,
            expected_route=None,
            expects_reply=False,
        )

    cast(Any, kernel)._session = SimpleNamespace(
        state=SessionState.ACTIVE, send_json=lambda *_a, **_k: None
    )
    assert (
        kernel._send_request_with_seq(
            2,
            "zone",
            "get_status",
            {"x": 1},
            pending=False,
            opaque=None,
            expected_route=None,
            expects_reply=False,
        )
        == 2
    )

    monkeypatch.setattr(kernel, "_send_request_with_seq", lambda *_a, **_k: 7)
    assert (
        kernel._send_request(
            "zone", "get_status", {"x": 1}, pending=False, opaque=None, expected_route=None
        )
        == 7
    )


def test_build_request_and_emit_and_envelopes(caplog: pytest.LogCaptureFixture) -> None:
    kernel = E27Kernel()
    msg = kernel._build_request_message(1, "zone", "get_status", "payload")
    assert msg["zone"]["get_status"] == "payload"

    kernel._log_outbound("control", "r_u_alive", msg)

    token = kernel.subscribe(lambda *_a, **_k: None, kinds={"other"})
    kernel.emit(
        ConnectionStateChanged(
            kind=ConnectionStateChanged.KIND,
            at=0.0,
            seq=None,
            classification="LOCAL",
            route=("__local__", "connection_state"),
            session_id=None,
            connected=True,
        ),
        ctx=kernel_mod.DispatchContext(
            kind=kernel_mod.MessageKind.UNKNOWN,
            seq=None,
            session_id=None,
            route=("__local__", "connection_state"),
            classification="LOCAL",
            response_match=None,
            raw_route=None,
        ),
    )
    kernel.unsubscribe(token)

    token = kernel.subscribe(lambda *_a, **_k: (_ for _ in ()).throw(RuntimeError("boom")))
    caplog.set_level("WARNING", logger="elke27_lib.kernel")
    kernel.emit(
        ConnectionStateChanged(
            kind=ConnectionStateChanged.KIND,
            at=0.0,
            seq=None,
            classification="LOCAL",
            route=("__local__", "connection_state"),
            session_id=None,
            connected=True,
        ),
        ctx=kernel_mod.DispatchContext(
            kind=kernel_mod.MessageKind.UNKNOWN,
            seq=None,
            session_id=None,
            route=("__local__", "connection_state"),
            classification="LOCAL",
            response_match=None,
            raw_route=None,
        ),
    )
    kernel.unsubscribe(token)

    assert (
        kernel._handle_dispatch_error_envelope(
            {},
            kernel_mod.DispatchContext(
                kind=kernel_mod.MessageKind.UNKNOWN,
                seq=None,
                session_id=None,
                route=("__local__", "connection_state"),
                classification="LOCAL",
                response_match=None,
                raw_route=None,
            ),
        )
        is False
    )
    assert (
        kernel._handle_dispatch_error_envelope(
            {"__error__": {"x": "y"}},
            kernel_mod.DispatchContext(
                kind=kernel_mod.MessageKind.UNKNOWN,
                seq=None,
                session_id=None,
                route=("__local__", "connection_state"),
                classification="LOCAL",
                response_match=None,
                raw_route=None,
            ),
        )
        is False
    )

    assert (
        kernel._handle_panel_error_envelope(
            {"error_code": "x"},
            kernel_mod.DispatchContext(
                kind=kernel_mod.MessageKind.UNKNOWN,
                seq=None,
                session_id=None,
                route=("__local__", "connection_state"),
                classification="LOCAL",
                response_match=None,
                raw_route=None,
            ),
        )
        is False
    )
    assert (
        kernel._handle_panel_error_envelope(
            {"error_code": "1", "error_message": 2},
            kernel_mod.DispatchContext(
                kind=kernel_mod.MessageKind.UNKNOWN,
                seq=None,
                session_id=None,
                route=("__local__", "connection_state"),
                classification="LOCAL",
                response_match=None,
                raw_route=None,
            ),
        )
        is True
    )



"""
Live E27 test: user enumeration via user_get_configured + user_get_attribs.

Run:
  source ~/elk-e27-env-vars.sh
  pytest -q -m live_e27 test/test_live_e27_users_download.py -s
"""

from __future__ import annotations

import logging
import os
from collections.abc import Callable
from typing import cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.const import E27ErrorCode
from test.helpers.error_codes import describe_error, extract_error_code
from test.helpers.internal import get_kernel, get_private, set_private
from test.helpers.payload_validation import assert_payload_shape

_LIVE_TIMEOUT_S = 15.0
LOG = logging.getLogger(__name__)


def _get_pin() -> int:
    pin_str = (os.environ.get("ELKE27_PIN") or "").strip()
    if not pin_str:
        pytest.skip("ELKE27_PIN not set; skipping auth-required test.")
    if not pin_str.isdigit():
        pytest.fail("ELKE27_PIN must be numeric.")
    pin_val = int(pin_str)
    if pin_val <= 0:
        pytest.fail("ELKE27_PIN must be a positive integer.")
    return pin_val


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_users_download(live_e27_client: Elke27Client) -> None:
    log_level = str(os.environ.get("LOG_LEVEL", "INFO") or "INFO").upper()
    logging.basicConfig(level=getattr(logging, log_level, logging.INFO), force=True)
    logging.getLogger("elke27_lib.session").setLevel(logging.DEBUG)

    pin = _get_pin()
    kernel = get_kernel(live_e27_client)
    original_on_message = cast(
        Callable[[dict[str, object]], object], get_private(kernel, "_on_message")
    )
    enable_raw_log = log_level == "DEBUG"
    if enable_raw_log:

        def _log_on_message(msg: dict[str, object]) -> object:
            LOG.debug("raw inbound msg=%r", msg)
            return original_on_message(msg)

        set_private(kernel, "_on_message", _log_on_message)
    try:
        auth_result = await live_e27_client.async_execute("control_authenticate", pin=pin)
        LOG.debug(
            "control.authenticate result ok=%s data=%r error=%s",
            auth_result.ok,
            auth_result.data,
            auth_result.error,
        )
        if not auth_result.ok:
            pytest.fail(f"control.authenticate failed: {describe_error(auth_result.error)}")
        assert_payload_shape("control_authenticate", auth_result.data)

        configured = await live_e27_client.async_execute(
            "user_get_configured", timeout_s=_LIVE_TIMEOUT_S
        )
        if not configured.ok:
            if extract_error_code(configured.error) == E27ErrorCode.ELKERR_NOT_READY:
                pytest.skip("Panel reported not-ready for user_get_configured.")
            pytest.fail(f"user_get_configured failed: {describe_error(configured.error)}")
        assert_payload_shape("user_get_configured", configured.data)

        user_ids = configured.data.get("users") if configured.data else None
        if not isinstance(user_ids, list):
            pytest.skip("No configured users reported.")

        user_ids_list = cast(list[object], user_ids)
        user_ids = [
            user_id for user_id in user_ids_list if isinstance(user_id, int) and user_id >= 1
        ]
        if not user_ids:
            pytest.skip("No configured users reported.")

        last_error = None
        for user_id in user_ids:
            auth_result = await live_e27_client.async_execute("control_authenticate", pin=pin)
            LOG.debug(
                "control.authenticate (per-user) ok=%s data=%r error=%s",
                auth_result.ok,
                auth_result.data,
                auth_result.error,
            )
            if not auth_result.ok:
                pytest.fail(f"control.authenticate failed: {describe_error(auth_result.error)}")
            assert_payload_shape("control_authenticate", auth_result.data)

            result = await live_e27_client.async_execute(
                "user_get_attribs",
                user_id=user_id,
                pin=str(pin),
                timeout_s=_LIVE_TIMEOUT_S,
            )
            if not result.ok:
                code = extract_error_code(result.error)
                if code == E27ErrorCode.ELKERR_NOT_READY:
                    pytest.skip("Panel reported not-ready for user_get_attribs.")
                if code == E27ErrorCode.ELKERR_NOAUTH:
                    pytest.fail(
                        f"user_get_attribs rejected after authenticate: {describe_error(result.error)}"
                    )
                last_error = result.error
                continue
            assert_payload_shape("user_get_attribs", result.data)
            if result.ok:
                user = live_e27_client.state.users.get(user_id)
                if user is not None:
                    return

        if last_error is not None:
            pytest.fail(
                f"User attribs unavailable after authenticate: {describe_error(last_error)}"
            )
        pytest.skip("User attribs unavailable; auth may be required.")
    finally:
        if enable_raw_log:
            set_private(kernel, "_on_message", original_on_message)



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from typing import cast

import pytest

from elke27_lib.client import Elke27Client
from test.helpers.internal import get_kernel, get_private, set_private


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_system_r_u_alive(live_e27_client: Elke27Client) -> None:
    kernel = get_kernel(live_e27_client)
    stop_keepalive = cast(Callable[[], None], get_private(kernel, "_stop_keepalive"))
    stop_keepalive()
    set_private(kernel, "_keepalive_interval_s", 1.0)
    set_private(kernel, "_keepalive_timeout_s", 2.0)
    set_private(kernel, "_keepalive_max_missed", 2)
    set_private(kernel, "_keepalive_enabled", True)

    fired = asyncio.Event()
    result_box: dict[str, bool] = {}
    original = cast(Callable[[], Awaitable[bool]], get_private(kernel, "_send_keepalive_request"))

    async def _wrapped_keepalive() -> bool:
        ok = await original()
        result_box["ok"] = ok
        fired.set()
        return ok

    set_private(kernel, "_send_keepalive_request", _wrapped_keepalive)
    start_keepalive = cast(Callable[[], None], get_private(kernel, "_start_keepalive"))
    start_keepalive()

    await asyncio.wait_for(fired.wait(), timeout=10.0)
    assert result_box.get("ok") is True



from __future__ import annotations

from elke27_lib.const import E27ErrorCode
from elke27_lib.dispatcher import DispatchContext, MessageKind
from elke27_lib.handlers.zone import make_zone_get_status_handler
from elke27_lib.states import PanelState


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, ctx: DispatchContext) -> None:
        self.events.append(evt)


def _ctx(route: tuple[str, str]) -> DispatchContext:
    return DispatchContext(
        kind=MessageKind.DIRECTED,
        seq=None,
        session_id=None,
        route=route,
        classification="RESPONSE",
    )


def test_zone_status_flags_all_false() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_zone_get_status_handler(state, emit, now=lambda: 123.0)

    msg = {
        "zone": {
            "get_status": {
                "zone_id": 1,
                "BYPASSED": False,
                "secure_state": "NORMAL",
                "low_batt": False,
                "error_code": E27ErrorCode.ELKERR_NONE,
            }
        }
    }
    assert handler(msg, _ctx(("zone", "get_status"))) is True
    zone = state.zones[1]
    assert zone.bypassed is False
    assert zone.trouble is False
    assert zone.tamper is False
    assert zone.alarm is False
    assert zone.low_battery is False


def test_zone_status_flags_multiple_true() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_zone_get_status_handler(state, emit, now=lambda: 123.0)

    msg = {
        "zone": {
            "get_status": {
                "zone_id": 2,
                "BYPASSED": True,
                "secure_state": "TAMPER",
                "trouble": True,
                "alarm": True,
                "low_batt": True,
                "error_code": E27ErrorCode.ELKERR_NONE,
            }
        }
    }
    assert handler(msg, _ctx(("zone", "get_status"))) is True
    zone = state.zones[2]
    assert zone.bypassed is True
    assert zone.trouble is True
    assert zone.tamper is True
    assert zone.alarm is True
    assert zone.low_battery is True



from __future__ import annotations

from elke27_lib.provisioning import ProvisioningManager


def test_provisioning_credentials_flow() -> None:
    mgr = ProvisioningManager()
    assert mgr.get_credentials() is None

    called: list[str] = []

    def _on_required(reason: str) -> None:
        called.append(reason)

    mgr.on_credentials_required = _on_required
    mgr.request_credentials("need creds")
    assert called == ["need creds"]

    mgr.supply_credentials("code", "pass")
    assert mgr.get_credentials() == ("code", "pass")

    mgr.clear_credentials()
    assert mgr.get_credentials() is None



from __future__ import annotations

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    ApiError,
    AuthorizationRequiredEvent,
    CsmSnapshotUpdated,
    KeypadConfiguredInventoryReady,
    TableCsmChanged,
)
from elke27_lib.handlers import keypad as keypad_handler
from elke27_lib.states import KeypadState, PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, _ctx: DispatchContext) -> None:
        self.events.append(evt)


def _any_event(spy: _EmitSpy, kind: type) -> bool:
    return any(isinstance(evt, kind) for evt in spy.events)


def test_keypad_get_configured_handler() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = keypad_handler.make_keypad_get_configured_handler(state, emit, now=lambda: 1.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"keypad": {}}, make_ctx()) is False

    msg = {"keypad": {"get_configured": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"keypad": {"get_configured": {"error_code": 9}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"keypad": {"get_configured": {"keypads": [1, 2], "block_id": 1, "block_count": 1}}}
    assert handler(msg, make_ctx()) is True
    assert state.inventory.configured_keypads == {1, 2}
    assert state.inventory.configured_keypads_complete is True
    assert _any_event(emit, KeypadConfiguredInventoryReady)
    assert state.panel.last_message_at == 1.0


def test_keypad_get_attribs_handler() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = keypad_handler.make_keypad_get_attribs_handler(state, emit, now=lambda: 2.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"keypad": {}}, make_ctx()) is False

    msg = {"keypad": {"get_attribs": {"error_code": 11008, "keypad_id": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"keypad": {"get_attribs": {"error_code": 7, "keypad_id": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    msg = {"keypad": {"get_attribs": {"keypad_id": 0}}}
    assert handler(msg, make_ctx()) is False

    msg = {
        "keypad": {
            "get_attribs": {
                "keypad_id": 1,
                "name": "  Front ",
                "area": 2,
                "zone_id": 3,
                "source_id": 4,
                "device_id": "dev",
                "flags": [1],
                "extra": "x",
            }
        }
    }
    assert handler(msg, make_ctx()) is True
    keypad = state.keypads[1]
    assert keypad.name == "Front"
    assert keypad.area == 2
    assert keypad.zone_id == 3
    assert keypad.source_id == 4
    assert keypad.device_id == "dev"
    assert keypad.flags == [1]
    assert keypad.fields["extra"] == "x"
    assert state.panel.last_message_at == 2.0


def test_keypad_get_table_info_handler_and_helpers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = keypad_handler.make_keypad_get_table_info_handler(state, emit, now=lambda: 3.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"keypad": {}}, make_ctx()) is False

    msg = {"keypad": {"get_table_info": {"error_code": 4}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"keypad": {"table_info": {"table_csm": "6"}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, TableCsmChanged)
    assert _any_event(emit, CsmSnapshotUpdated)

    assert keypad_handler._extract_configured_ids({"keypads": [1, 2, "x"]}) == {1, 2}
    assert keypad_handler._extract_configured_ids({"keypad_ids": [3]}) == {3}
    assert keypad_handler._extract_configured_ids({"configured_keypads": [4]}) == {4}
    assert keypad_handler._extract_configured_ids({"configured_keypad_ids": [5]}) == {5}
    assert keypad_handler._extract_configured_ids({"none": []}) == set()

    assert keypad_handler._normalize_name(None) is None
    assert keypad_handler._normalize_name("   ") is None
    assert keypad_handler._normalize_name(" Name ") == "Name"

    keypad_state = KeypadState(keypad_id=1)
    changed: set[str] = set()
    keypad_handler._apply_keypad_attribs(
        keypad_state,
        {
            "name": " Bob ",
            "area": 1,
            "zone_id": 2,
            "source_id": 3,
            "device_id": "d",
            "flags": [1],
            "extra": "x",
        },
        changed,
    )
    assert keypad_state.name == "Bob"
    assert keypad_state.area == 1
    assert keypad_state.zone_id == 2
    assert keypad_state.source_id == 3
    assert keypad_state.device_id == "d"
    assert keypad_state.flags == [1]
    assert keypad_state.fields["extra"] == "x"

    assert keypad_handler._extract_table_csm({"table_csm": True}, domain="keypad") is None
    assert keypad_handler._extract_table_csm({"table_csm": 2}, domain="keypad") == 2
    assert keypad_handler._extract_table_csm({"table_csm": 2.0}, domain="keypad") == 2
    assert keypad_handler._extract_table_csm({"table_csm": "3"}, domain="keypad") == 3
    assert keypad_handler._extract_table_csm({"table_csm": "bad"}, domain="keypad") is None
    assert keypad_handler._extract_table_csm({"other": 1}, domain="keypad") is None



from collections.abc import Callable, Iterable, Iterator
from typing import cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    AreaConfiguredInventoryReady,
    AreaStatusUpdated,
    Event,
    OutputConfiguredInventoryReady,
    OutputStatusUpdated,
    ZoneConfiguredInventoryReady,
    ZoneStatusUpdated,
)
from elke27_lib.kernel import E27Kernel
from elke27_lib.states import PanelState
from test.helpers.internal import get_private


class _FakeKernel:
    state: PanelState
    requests: list[tuple[tuple[str, str], dict[str, object]]]

    def __init__(self) -> None:
        self.state = PanelState()
        self.state.panel.session_id = 1
        self.state.table_info_by_domain = {
            "area": {"table_elements": 1},
            "zone": {"table_elements": 1},
            "output": {"table_elements": 1},
            "tstat": {"table_elements": 1},
        }
        self.requests = []

    @property
    def ready(self) -> bool:
        return True

    def request(self, route: tuple[str, str], **kwargs: object) -> None:
        self.requests.append((route, dict(kwargs)))

    def subscribe(
        self, _callback: Callable[[Event], None], _kinds: Iterable[str] | None = None
    ) -> int:
        return 1

    def drain_events(self) -> list[Event]:
        return []

    def iter_events(self) -> Iterator[Event]:
        return iter(())


def _inventory_ready_event(evt_cls: type[Event]) -> Event:
    kind_attr = "KIND"
    kind = cast(str, getattr(evt_cls, kind_attr))
    return evt_cls(
        kind=kind,
        at=UNSET_AT,
        seq=UNSET_SEQ,
        classification=UNSET_CLASSIFICATION,
        route=UNSET_ROUTE,
        session_id=UNSET_SESSION_ID,
    )


@pytest.mark.asyncio
async def test_wait_ready_requires_inventory_and_status():
    kernel = _FakeKernel()
    client = Elke27Client(kernel=cast(E27Kernel, cast(object, kernel)))

    assert await client.wait_ready(timeout_s=0.01) is False

    kernel.state.inventory.configured_areas = {1}
    kernel.state.inventory.configured_zones = {1}
    kernel.state.inventory.configured_outputs = {1}

    handle_event = get_private(client, "_handle_kernel_event")
    handle_event(_inventory_ready_event(AreaConfiguredInventoryReady))
    handle_event(_inventory_ready_event(ZoneConfiguredInventoryReady))
    handle_event(_inventory_ready_event(OutputConfiguredInventoryReady))

    assert await client.wait_ready(timeout_s=0.01) is False

    kernel.state.get_or_create_area(1)
    kernel.state.get_or_create_zone(1)
    kernel.state.get_or_create_output(1)

    handle_event(
        AreaStatusUpdated(
            kind=AreaStatusUpdated.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            area_id=1,
            changed_fields=("arm_state",),
        )
    )
    handle_event(
        ZoneStatusUpdated(
            kind=ZoneStatusUpdated.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            zone_id=1,
            changed_fields=("violated",),
        )
    )
    handle_event(
        OutputStatusUpdated(
            kind=OutputStatusUpdated.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            output_id=1,
            status="ON",
            on=True,
        )
    )

    assert await client.wait_ready(timeout_s=0.1) is True


def test_bootstrap_populates_snapshot_maps():
    kernel = _FakeKernel()
    client = Elke27Client(kernel=cast(E27Kernel, cast(object, kernel)))

    kernel.state.inventory.configured_areas = {1}
    kernel.state.inventory.configured_zones = {1}
    kernel.state.inventory.configured_outputs = {1}

    handle_event = get_private(client, "_handle_kernel_event")
    handle_event(_inventory_ready_event(AreaConfiguredInventoryReady))
    handle_event(_inventory_ready_event(ZoneConfiguredInventoryReady))
    handle_event(_inventory_ready_event(OutputConfiguredInventoryReady))

    assert (("area", "get_status"), {"area_id": 1}) in kernel.requests
    assert (("zone", "get_status"), {"zone_id": 1}) in kernel.requests
    assert (("output", "get_status"), {"output_id": 1}) in kernel.requests

    kernel.state.get_or_create_area(1)
    kernel.state.get_or_create_zone(1)
    kernel.state.get_or_create_output(1)

    handle_event(
        AreaStatusUpdated(
            kind=AreaStatusUpdated.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            area_id=1,
            changed_fields=("arm_state",),
        )
    )
    handle_event(
        ZoneStatusUpdated(
            kind=ZoneStatusUpdated.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            zone_id=1,
            changed_fields=("violated",),
        )
    )
    handle_event(
        OutputStatusUpdated(
            kind=OutputStatusUpdated.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            output_id=1,
            status="ON",
            on=True,
        )
    )

    assert client.snapshot.areas
    assert client.snapshot.zones
    assert client.snapshot.outputs
